#!/usr/bin/env python

import sys

from ..log import *
from ..niche_net import gen_original_data
from ..optparser import (opt_preprocessing_validate,
                         prepare_preprocessing_optparser)
from ..run.processes import load_parameters, niche_network_construct
from ..utils import write_version_info


# ------------------------------------
# Main Function
# ------------------------------------
def main() -> None:
    """
    main function
    Input data files information should be stored in a YAML file.
    """

    # write version information
    write_version_info()

    # load parameters
    options = load_parameters(opt_validate_func=opt_preprocessing_validate,
                              prepare_optparser_func=prepare_preprocessing_optparser)

    # load original data
    ori_data_df = gen_original_data(options=options)

    # ----- Niche Network Construct -----
    niche_network_construct(options=options, ori_data_df=ori_data_df)


# ------------------------------------
# Program running
# ------------------------------------
if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        sys.stderr.write("User interrupts me! ;-) See you ^.^!\n")
        sys.exit(0)



import os
from optparse import Values
from typing import Any, Dict, List

import numpy as np
import pandas as pd
import yaml
from scipy.linalg import cholesky, eigh
from scipy.sparse import csr_matrix, save_npz
from scipy.spatial import cKDTree, distance

from ..log import *


def gauss_dist_1d(dist: np.ndarray, n_local: int) -> float:
    """
    Compute gaussian affinity between two cells (a cell and its KNN)
    :param dist_use: knn spatial distance to be used
    :param n_local: index of distance used for normalization
    :return: gaussian distance
    """
    return np.exp(-(dist / dist[n_local])**2)


def construct_niche_network_sample(options: Values, sample_data_df: pd.DataFrame, sample_name: str) -> None:
    """
    Construct niche network for a sample
    :param options: Values, options
    :param sample_data_df: pd.DataFrame, sample data
    :param sample_name: str, sample name
    :return: None

    1) get coordinates and save it.
    2) save the celltype information
    3) build KDTree
        1. save edge index file
        2. calculate weight matrix
        3. calculate cell type composition and save it
    """

    info(f'Constructing niche network for sample: {sample_name}.')

    n_local = options.n_local
    N = sample_data_df.shape[0]

    # get coordinates
    # TODO: support 3D coordinates
    coord_df = sample_data_df[['Cell_ID', 'x', 'y']]
    coord_df.to_csv(f'{options.preprocessing_dir}/{sample_name}_Coordinates.csv', index=False)

    # build KDTree
    coordinates = sample_data_df[['x', 'y']].values
    kdtree = cKDTree(data=coordinates)
    dis_matrix, indices_matrix = kdtree.query(x=coordinates,
                                              k=np.max([options.n_neighbors, options.n_local]) + 1)  # include self
    np.savetxt(f'{options.preprocessing_dir}/{sample_name}_NeighborIndicesMatrix.csv.gz', indices_matrix,
               delimiter=',')  # save indices matrix

    # save edge index file
    # 1) convert edge index to csr_matrix
    # 2) make it bidirectional
    # 3) convert it to edge index back
    # 4) save it
    src_indices = np.repeat(np.arange(coordinates.shape[0]), options.n_neighbors)
    dst_indices = indices_matrix[:, 1:options.n_neighbors + 1].flatten()  # remove self, N x k, #niche x #cell
    adj_matrix = csr_matrix((np.ones(dst_indices.shape[0]), (src_indices, dst_indices)),
                            shape=(N, N))  # convert to csr_matrix
    adj_matrix = adj_matrix + adj_matrix.transpose()  # make it bidirectional
    edge_index = np.argwhere(adj_matrix > 0)  # convert it to edge index back
    edge_index_file = f'{options.preprocessing_dir}/{sample_name}_EdgeIndex.csv.gz'
    np.savetxt(edge_index_file, edge_index, delimiter=',', fmt='%d')

    # calculate niche_weight_matrix and normalize it using self node and n_local-th neighbor using a gaussian kernel
    # calculate cell_to_niche_matrix
    niche_weight_matrix = np.apply_along_axis(func1d=gauss_dist_1d, axis=1, arr=dis_matrix,
                                              n_local=n_local)[:, :options.n_neighbors +
                                                               1]  # N x (k + 1), #niche x #cell
    src_indices = np.repeat(np.arange(coordinates.shape[0]), options.n_neighbors + 1)
    dst_indices = indices_matrix[:, :options.n_neighbors + 1].flatten()  # include self, N x (k + 1), #niche x #cell
    niche_weight_matrix_csr = csr_matrix((niche_weight_matrix.flatten(), (src_indices, dst_indices)),
                                         shape=(N, N))  # convert to csr_matrix, N x N, #niche x #cell
    save_npz(file=f'{options.preprocessing_dir}/{sample_name}_NicheWeightMatrix.npz',
             matrix=niche_weight_matrix_csr)  # save weight matrix
    cell_to_niche_matrix = niche_weight_matrix_csr / niche_weight_matrix_csr.sum(axis=1)  # N x N, #niche x #cell

    # calculate cell type composition
    if options.decomposition_cell_type_composition_input is None:
        one_hot_matrix = np.zeros(shape=(N, sample_data_df['Cell_Type'].cat.categories.shape[0]))  # N (#cell) x #cell_type
        one_hot_matrix[np.arange(N), sample_data_df.Cell_Type.cat.codes.values] = 1
        cell_type_composition = cell_to_niche_matrix @ one_hot_matrix  # N (#niche) x #cell_type
    else: # use the provided cell type composition, spot-based data
        decomposition_cell_type_composition_df = pd.read_csv(options.decomposition_cell_type_composition_input,
                                                                 header=0,
                                                                 index_col=0)  # N (#cell) x #cell_type
        cell_type_composition = cell_to_niche_matrix @ decomposition_cell_type_composition_df.values  # N (#niche) x #cell_type

    # save cell type composition
    np.savetxt(f'{options.preprocessing_dir}/{sample_name}_CellTypeComposition.csv.gz',
               cell_type_composition,
               delimiter=',')


def construct_niche_network(options: Values, ori_data_df: pd.DataFrame) -> None:
    """
    Construct niche network
    :param ori_data_df: pd.DataFrame, original data
    :return: None
    """

    # get samples
    samples = ori_data_df['Sample'].unique()

    # construct niche network for each sample
    for sample in samples:
        sample_data_df = ori_data_df[ori_data_df['Sample'] == sample]
        construct_niche_network_sample(options=options, sample_data_df=sample_data_df, sample_name=sample)


def gen_samples_yaml(options: Values, ori_data_df: pd.DataFrame) -> None:
    """
    Generate samples.yaml
    :param ori_data_df: pd.DataFrame, original data
    :return: None
    """

    info('Generating samples.yaml file.')

    data: Dict[str, List[Any]] = {'Data': []}
    for sample in ori_data_df['Sample'].unique():
        data['Data'].append({
            'Name': f'{sample}',
            'Coordinates': f'{sample}_Coordinates.csv',
            'EdgeIndex': f'{sample}_EdgeIndex.csv.gz',
            'Features': f'{sample}_CellTypeComposition.csv.gz',
            'NicheWeightMatrix': f'{sample}_NicheWeightMatrix.npz',
            'NeighborIndicesMatrix': f'{sample}_NeighborIndicesMatrix.csv.gz'
        })

    yaml_file = f'{options.preprocessing_dir}/samples.yaml'
    with open(yaml_file, 'w') as fhd:
        yaml.dump(data, fhd)


def get_embedding_columns(ori_data_df: pd.DataFrame) -> List[str]:
    """
    Get embedding columns from the original data
    :param ori_data_df: pd.DataFrame, original data
    :return: List[str], embedding columns
    """
    embedding_start_column = [x for x in ori_data_df.columns if x.startswith('Embedding_')]
    embedding_columns = []
    i = 0
    while True:
        i += 1
        if f'Embedding_{i}' in embedding_start_column:
            embedding_columns.append(f'Embedding_{i}')
        else:
            break

    return embedding_columns


def ct_coding_adjust(options: Values, ori_data_df: pd.DataFrame):
    """
    Adjust the cell type coding according to embeddings

    1) check the embedding info in the original data
    2) calculate embedding postion for each cell type
    3) calculate distance between each cell type
    4) calculate the M
    5) got the new basis for cell type coding
    6) adjust the cell type coding

    :param options: Values, options
    :param ori_data_df: pd.DataFrame, original data
    :return: None
    """

    if options.decomposition_cell_type_composition_input is None:
        # check the embedding info in the original data
        embedding_columns = get_embedding_columns(ori_data_df)
        if len(embedding_columns) < 2:
            warning('At least two (Embedding_1 and Embedding_2) should be in the original data. Skip the adjustment.')
            return

        # calculate embedding postion for each cell type
        ct_embedding = ori_data_df[embedding_columns + ['Cell_Type']].groupby('Cell_Type').mean()
        # calculate distance between each cell type
        raw_distance = distance.cdist(ct_embedding[embedding_columns].values, ct_embedding[embedding_columns].values,
                                  'euclidean')
    
    else:  # spot-based data
        ct_embedding = np.loadtxt(f'{options.preprocessing_dir}/PCA_embedding.csv', delimiter=',')
        raw_distance = distance.cdist(ct_embedding, ct_embedding, 'euclidean')

    median_distance = np.median(raw_distance[np.triu_indices(raw_distance.shape[0], k=1)])
    info(f'Median distance between cell types: {median_distance}')

    # calculate the M
    M = np.exp(-(raw_distance / (options.sigma * median_distance)) **2)

    # got the new basis for cell type coding
    eig_val, eig_vec = eigh(M)
    if np.any(eig_val < 0):
        warning('Negative eigenvalues found. Skip the adjustment.')
        return
    L = cholesky(M)

    # adjust the cell type coding
    for sample in ori_data_df['Sample'].unique():
        feat_file = f'{options.preprocessing_dir}/{sample}_CellTypeComposition.csv.gz'
        ctc_raw = np.loadtxt(feat_file, delimiter=',')
        ctc_new = ctc_raw @ L.T
        os.rename(feat_file, f'{options.preprocessing_dir}/{sample}_Raw_CellTypeComposition.csv.gz')
        np.savetxt(feat_file, ctc_new, delimiter=',')



import os
from optparse import Values
from typing import Dict, Union

import numpy as np
import pandas as pd
from pandas import DataFrame

from ..log import info, warning
from ..utils import get_rel_params, read_yaml_file


# ----------------------------
# Misc Functions
# ----------------------------
def load_loss_record_data(options) -> Union[Dict, None]:
    """
    Parse the log file and save the training loss to a csv file.
    Args:
        options: Values, the options from optparse
    Returns:
        Dict: {'loss_df': pd.DataFrame, 'loss_dict': Dict}
        loss_df is the training loss dataframe
        loss_dict is the final loss dictionary

    loss record example:
        16:31:47 --- INFO: epoch: 1, batch: 1, loss: 8.554190635681152, spectral_loss: -6.64739809508319e-06, cluster_loss: 0.06959901750087738, feat_similarity_loss: 8.484598159790039
    """
    loss, loss_name = [], []  # type: ignore
    if options.log is None:
        return None
    with open(options.log, 'r') as fhd:
        for line in fhd:
            # train loss record
            if 'INFO' in line and 'epoch' in line and 'loss' in line:
                loss_ = []
                loss_name = []
                line: list[str] = line.strip().split()  # type: ignore
                init_index = line.index('loss:')  # get the first loss index
                # get loss name & losss
                for i in range(init_index, len(line), 2):  # loss name
                    loss_name.append(line[i].strip(':'))
                for i in range(init_index + 1, len(line), 2):  # loss value
                    loss_.append(float(line[i].strip(',')))
                # insert epoch and batch information
                loss_.insert(0, int(line[init_index - 1].strip(',')))  # batch index
                loss_.insert(0, int(line[init_index - 3].strip(',')))  # epoch index
                loss.append(loss_)
            # eval loss record
            elif 'INFO' in line and 'Evaluate loss' in line:
                line = line.strip().split(', ', 1)  # type: ignore
                final_loss_dict: Dict[str, float] = eval(line[1])

    # loss name edit
    loss_name[0] = 'total_loss'
    loss_name = ['Epoch', 'Batch'] + loss_name[:]

    # save loss to csv
    loss_df = pd.DataFrame(loss, columns=loss_name)
    if options.output is not None:
        loss_df.to_csv(f'{options.output}/train_loss.csv', index=False)

    # calculate epoch loss
    epoch_loss_columns = [col for col in loss_df.columns if 'loss' in col]
    epoch_loss_columns.insert(0, 'Epoch')
    epoch_loss_df = loss_df[epoch_loss_columns].groupby('Epoch').mean()
    if options.output is not None:
        epoch_loss_df.to_csv(f'{options.output}/train_loss_epoch.csv')

    return {'loss_df': epoch_loss_df, 'loss_dict': final_loss_dict}


def load_niche_cluster_connectivity(options: Values) -> np.ndarray:
    """
    Load the niche cluster connectivity from the output of GNN.
    Args:
        options: Values, the options from optparse
    Returns:
        np.ndarray: the niche cluster connectivity
    """
    niche_cluster_conn_file = f'{options.GNN_dir}/consolidate_out_adj.csv.gz'
    if not os.path.isfile(niche_cluster_conn_file):
        niche_cluster_conn_file = f'{options.GNN_dir}/consolidate_out_adj.csv'
    if not os.path.isfile(niche_cluster_conn_file):  # skip if file not exist
        raise FileNotFoundError(f"Cannot find niche cluster connectivity file: {niche_cluster_conn_file}.")

    return np.loadtxt(f'{niche_cluster_conn_file}', delimiter=',')


def load_niche_cluster_score(options: Values) -> np.ndarray:
    """
    Load the niche cluster score from the output of NT score.
    Args:
        options: Values, the options from optparse
    Returns:
        np.ndarray: the niche cluster score
    """
    niche_cluster_score_file = f'{options.NTScore_dir}/niche_cluster_score.csv.gz'
    if not os.path.isfile(niche_cluster_score_file):
        niche_cluster_score_file = f'{options.NTScore_dir}/niche_cluster_score.csv'
    if not os.path.isfile(niche_cluster_score_file):  # skip if file not exist
        raise FileNotFoundError(f"Cannot find niche cluster score file: {niche_cluster_score_file}.")

    return np.loadtxt(f'{niche_cluster_score_file}', delimiter=',')


def load_niche_level_niche_cluster_assign(options: Values) -> pd.DataFrame:
    """
    Load the niche cluster assignment for each niche level.
    Args:
        options: Values, the options from optparse
    Returns:
        pd.DataFrame: the niche cluster assignment for each niche level
    """

    niche_level_niche_cluster_assign_file = f'{options.GNN_dir}/niche_level_niche_cluster.csv.gz'
    if not os.path.isfile(niche_level_niche_cluster_assign_file):
        niche_level_niche_cluster_assign_file = f'{options.GNN_dir}/niche_level_niche_cluster.csv'
    if not os.path.isfile(niche_level_niche_cluster_assign_file):  # skip if file not exist
        raise FileNotFoundError(
            f"Cannot find niche level niche cluster assign file: {niche_level_niche_cluster_assign_file}.")

    return pd.read_csv(niche_level_niche_cluster_assign_file, index_col=0)


def load_cell_level_niche_cluster_assign(options: Values) -> pd.DataFrame:
    """
    Load the niche cluster assignment for each cell level.
    Args:
        options: Values, the options from optparse
    Returns:
        pd.DataFrame: the niche cluster assignment for each cell level
    """

    cell_level_niche_cluster_assign_file = f'{options.GNN_dir}/cell_level_niche_cluster.csv.gz'
    if not os.path.isfile(cell_level_niche_cluster_assign_file):
        cell_level_niche_cluster_assign_file = f'{options.GNN_dir}/cell_level_niche_cluster.csv'
    if not os.path.isfile(cell_level_niche_cluster_assign_file):  # skip if file not exist
        raise FileNotFoundError(
            f"Cannot find cell level niche cluster assign file: {cell_level_niche_cluster_assign_file}.")

    return pd.read_csv(cell_level_niche_cluster_assign_file, index_col=0)


def load_niche_level_max_niche_cluster(options: Values) -> pd.DataFrame:
    """
    Load the max niche cluster assignment for each niche level.
    Args:
        options: Values, the options from optparse
    Returns:
        pd.DataFrame: the max niche cluster assignment for each niche level
    """

    niche_level_max_niche_cluster_file = f'{options.GNN_dir}/niche_level_max_niche_cluster.csv.gz'
    if not os.path.isfile(niche_level_max_niche_cluster_file):
        niche_level_max_niche_cluster_file = f'{options.GNN_dir}/niche_level_max_niche_cluster.csv'
    if not os.path.isfile(niche_level_max_niche_cluster_file):  # skip if file not exist
        raise FileNotFoundError(
            f"Cannot find niche level max niche cluster file: {niche_level_max_niche_cluster_file}.")

    return pd.read_csv(niche_level_max_niche_cluster_file, index_col=0)


def load_cell_level_max_niche_cluster(options: Values) -> pd.DataFrame:
    """
    Load the max niche cluster assignment for each cell level.
    Args:
        options: Values, the options from optparse
    Returns:
        pd.DataFrame: the max niche cluster assignment for each cell level
    """

    cell_level_max_niche_cluster_file = f'{options.GNN_dir}/cell_level_max_niche_cluster.csv.gz'
    if not os.path.isfile(cell_level_max_niche_cluster_file):
        cell_level_max_niche_cluster_file = f'{options.GNN_dir}/cell_level_max_niche_cluster.csv'
    if not os.path.isfile(cell_level_max_niche_cluster_file):  # skip if file not exist
        raise FileNotFoundError(f"Cannot find cell level max niche cluster file: {cell_level_max_niche_cluster_file}.")

    return pd.read_csv(cell_level_max_niche_cluster_file, index_col=0)


# ----------------------------
# Classes
# ----------------------------


class AnaData:
    """
    Class to store the data for analysis
    This class have the following attributes:
    - options: Values, the options from optparse
    - rel_params: Dict, the relative paths for params
    - cell_id: pd.DataFrame, the original Cell ID and Cell Type
    - train_loss: Dict, the training loss
    - cell_type_codes: pd.DataFrame, the cell type codes
    - cell_type_composition: pd.DataFrame, the cell type composition
    - NT_score: pd.DataFrame, the NT score
    - niche_cluster_connectivity: np.ndarray, the niche cluster connectivity
    - niche_cluster_score: np.ndarray, the niche cluster score
    - niche_level_niche_cluster_assign: pd.DataFrame, the niche cluster assignment for each niche level
    - cell_level_niche_cluster_assign: pd.DataFrame, the niche cluster assignment for each cell level
    - niche_level_max_niche_cluster: pd.DataFrame, the max niche cluster assignment for each niche level
    - cell_level_max_niche_cluster: pd.DataFrame, the max niche cluster assignment for each cell level
    """

    def __init__(self, options: Values) -> None:
        """
        Initialize the class with the options"""

        # save options
        self.options = options
        # get real path
        params = read_yaml_file(f'{options.preprocessing_dir}/samples.yaml')
        self.rel_params = get_rel_params(options, params)
        # save the original Cell ID
        self.cell_id = pd.read_csv(f'{options.preprocessing_dir}/meta_data.csv',
                                   usecols=['Cell_ID']).set_index('Cell_ID', inplace=False)
        self.meta_data = pd.read_csv(f'{options.preprocessing_dir}/meta_data.csv').set_index('Cell_ID', inplace=False)
        if 'Cell_Type' in self.meta_data.columns:
            self.cell_id['Cell_Type'] = self.meta_data['Cell_Type'] = self.meta_data['Cell_Type'].astype('category')

    @property
    def train_loss(self):
        if not hasattr(self, '_train_loss'):
            self._train_loss = load_loss_record_data(self.options)
        return self._train_loss

    @property
    def cell_type_codes(self) -> pd.DataFrame:
        if not hasattr(self, '_cell_type_codes'):
            self._cell_type_codes = pd.read_csv(f'{self.options.preprocessing_dir}/cell_type_code.csv', index_col=0)
        return self._cell_type_codes

    def _load_cell_type_composition_and_NT_score(self) -> None:
        data_df, data_2_df = pd.DataFrame(), pd.DataFrame()
        for sample in self.rel_params['Data']:
            # raw cell type composition
            cell_type_composition_df = pd.read_csv(sample['Features'], header=None)
            cell_type_composition_df.columns = self.cell_type_codes.loc[np.arange(cell_type_composition_df.shape[1]),
                                                                        'Cell_Type'].tolist()  # type: ignore
            NTScore_df = pd.read_csv(f'{self.options.NTScore_dir}/{sample["Name"]}_NTScore.csv.gz', index_col=0)
            sample_df = pd.concat([NTScore_df.reset_index(drop=True), cell_type_composition_df], axis=1)
            sample_df.index = NTScore_df.index
            sample_df['sample'] = [sample["Name"]] * sample_df.shape[0]
            data_df = pd.concat([data_df, sample_df])
            # adjust cell type composition
            if not self.options.embedding_adjust:
                continue
            feature_file = f"{sample['Features'][:-27]}_Raw_CellTypeComposition.csv.gz"
            if not os.path.isfile(feature_file):
                continue
            cell_type_composition_df = pd.read_csv(feature_file, header=None)
            cell_type_composition_df.columns = self.cell_type_codes.loc[np.arange(cell_type_composition_df.shape[1]),
                                                                        'Cell_Type'].tolist()  # type: ignore
            sample_df = pd.concat([NTScore_df.reset_index(drop=True), cell_type_composition_df], axis=1)
            sample_df.index = NTScore_df.index
            sample_df['sample'] = [sample["Name"]] * sample_df.shape[0]
            data_2_df = pd.concat([data_2_df, sample_df])

        data_df = self.cell_id.join(data_df)
        self._cell_type_composition = data_df[self.cell_type_codes['Cell_Type'].tolist() + ['sample', 'x', 'y']]
        self._NT_score = data_df[['sample', 'x', 'y', 'Niche_NTScore', 'Cell_NTScore']]
        if not data_2_df.empty:
            data_2_df = self.cell_id.join(data_2_df)
            self._adjust_cell_type_composition = data_2_df[self.cell_type_codes['Cell_Type'].tolist() +
                                                           ['sample', 'x', 'y']]

    @property
    def cell_type_composition(self) -> DataFrame:
        if not hasattr(self, '_cell_type_composition'):
            self._load_cell_type_composition_and_NT_score()
        return self._cell_type_composition

    @property
    def adjust_cell_type_composition(self) -> DataFrame:
        if not self.options.embedding_adjust:
            warning('The embedding adjust is not enabled. Skip the adjust cell type composition loading.')
        if not hasattr(self, '_adjust_cell_type_composition'):
            self._load_cell_type_composition_and_NT_score()
        return self._adjust_cell_type_composition

    @property
    def NT_score(self) -> DataFrame:
        if not hasattr(self, '_NT_score'):
            self._load_cell_type_composition_and_NT_score()
        return self._NT_score

    @property
    def niche_cluster_connectivity(self) -> np.ndarray:
        if not hasattr(self, '_niche_cluster_connectivity'):
            # FileNotFoundError will be raised if the file does not exist
            self._niche_cluster_connectivity = load_niche_cluster_connectivity(self.options)
        return self._niche_cluster_connectivity

    @property
    def niche_cluster_score(self) -> np.ndarray:
        if not hasattr(self, '_niche_cluster_score'):
            # FileNotFoundError will be raised if the file does not exist
            self._niche_cluster_score = load_niche_cluster_score(self.options)
        return self._niche_cluster_score

    @property
    def niche_level_niche_cluster_assign(self) -> pd.DataFrame:
        if not hasattr(self, '_niche_level_niche_cluster_assign'):
            self._niche_level_niche_cluster_assign = load_niche_level_niche_cluster_assign(self.options)
            try:
                self._niche_level_niche_cluster_assign = self._niche_level_niche_cluster_assign[self.cell_id.index]
            except:
                pass
        return self._niche_level_niche_cluster_assign

    @property
    def cell_level_niche_cluster_assign(self) -> pd.DataFrame:
        if not hasattr(self, '_cell_level_niche_cluster_assign'):
            self._cell_level_niche_cluster_assign = load_cell_level_niche_cluster_assign(self.options)
            try:
                self._cell_level_niche_cluster_assign = self._cell_level_niche_cluster_assign[self.cell_id.index]
            except:
                pass
        return self._cell_level_niche_cluster_assign

    @property
    def niche_level_max_niche_cluster(self) -> pd.DataFrame:
        if not hasattr(self, '_niche_level_max_niche_cluster'):
            self._niche_level_max_niche_cluster = load_niche_level_max_niche_cluster(self.options)
            try:
                self._niche_level_max_niche_cluster = self._niche_level_max_niche_cluster[self.cell_id.index]
            except:
                pass
        return self._niche_level_max_niche_cluster

    @property
    def cell_level_max_niche_cluster(self) -> pd.DataFrame:
        if not hasattr(self, '_cell_level_max_niche_cluster'):
            self._cell_level_max_niche_cluster = load_cell_level_max_niche_cluster(self.options)
            try:
                self._cell_level_max_niche_cluster = self._cell_level_max_niche_cluster[self.cell_id.index]
            except:
                pass
        return self._cell_level_max_niche_cluster



import itertools
import os
from optparse import Values
from typing import Dict, List, Tuple

import numpy as np
import pandas as pd
from numpy import ndarray
from scipy.sparse import load_npz

from ..data import SpatailOmicsDataset
from ..log import error, info


def load_consolidate_data(options: Values) -> Tuple[ndarray, ndarray]:
    """
    Load consolidate s_array and out_adj_array
    :param options: Values, options
    :return: Tuple[ndarray, ndarray], the consolidate s_array and out_adj_array
    """

    info('Loading consolidate s_array and out_adj_array...')

    if not os.path.exists(f'{options.GNN_dir}/consolidate_s.csv.gz') or not os.path.exists(
            f'{options.GNN_dir}/consolidate_out_adj.csv.gz'):
        error(f'consolidate_s.csv.gz or consolidate_out_adj.csv.gz does not exist in {options.GNN_dir} directory.')
    consolidate_s_array = np.loadtxt(fname=f'{options.GNN_dir}/consolidate_s.csv.gz', delimiter=',')
    consolidate_out_adj_array = np.loadtxt(fname=f'{options.GNN_dir}/consolidate_out_adj.csv.gz', delimiter=',')

    return consolidate_s_array, consolidate_out_adj_array


def get_niche_trajectory_path(niche_adj_matrix: ndarray) -> List[int]:
    """
    Find niche level trajectory with maximum connectivity using Brute Force
    :param adj_matrix: non-negative ndarray, adjacency matrix of the graph
    :return: List[int], the niche trajectory
    """

    info('Finding niche trajectory with maximum connectivity using Brute Force.')

    max_connectivity = float('-inf')
    niche_trajectory_path = []
    for path in itertools.permutations(range(len(niche_adj_matrix))):
        connectivity = 0
        for i in range(len(path) - 1):
            connectivity += niche_adj_matrix[path[i], path[i + 1]]
        if connectivity > max_connectivity:
            max_connectivity = connectivity
            niche_trajectory_path = list(path)

    return niche_trajectory_path


def trajectory_path_to_NC_score(options: Values, niche_trajectory_path: List[int],
                                niche_clustering_sum: ndarray) -> ndarray:
    """
    Convert niche cluster trajectory path to NTScore
    :param niche_trajectory_path: List[int], the niche trajectory path
    :return: ndarray, the NTScore
    """

    info('Calculating NTScore for each niche cluster based on the trajectory path.')

    niche_NT_score = np.zeros(len(niche_trajectory_path))
    if options.equal_space:
        values = np.linspace(0, 1, len(niche_trajectory_path))
        for i, index in enumerate(niche_trajectory_path):
            # debug(f'i: {i}, index: {index}')
            niche_NT_score[index] = values[i]
        return niche_NT_score
    else:
        sum = 0
        for i, index in enumerate(niche_trajectory_path):
            sum += niche_clustering_sum[index]
            niche_NT_score[index] = sum
        niche_NT_score -= niche_NT_score.min()
        niche_NT_score /= niche_NT_score.max()
        return niche_NT_score


def get_niche_NTScore(options: Values, niche_cluster_loading: ndarray,
                      niche_adj_matrix: ndarray) -> Tuple[ndarray, ndarray]:
    """
    Get niche-level niche trajectory and cell-level niche trajectory
    :param niche_cluster_loading: ndarray, the loading of cell x niche clusters
    :param adj_matrix: ndarray, the adjacency matrix of the graph
    :return: Tuple[ndarray, ndarray], the niche-level niche trajectory and cell-level niche trajectory
    """

    info('Calculating NTScore for each niche.')

    niche_trajectory_path = get_niche_trajectory_path(niche_adj_matrix=niche_adj_matrix)

    niche_clustering_sum = niche_cluster_loading.sum(axis=0)
    niche_cluster_score = trajectory_path_to_NC_score(options=options,
                                                      niche_trajectory_path=niche_trajectory_path,
                                                      niche_clustering_sum=niche_clustering_sum)
    niche_level_NTScore = niche_cluster_loading @ niche_cluster_score
    return niche_cluster_score, niche_level_NTScore


def niche_to_cell_NTScore(dataset: SpatailOmicsDataset, rel_params: Dict,
                          niche_level_NTScore: ndarray) -> Tuple[ndarray, Dict[str, ndarray], Dict[str, ndarray]]:
    """
    get cell-level NTScore
    :param dataset: SpatailOmicsDataset, dataset
    :param rel_params: Dict, relative paths
    :param niche_level_NTScore: ndarray, niche-level NTScore
    :return: Tuple[ndarray, Dict[str, ndarray], Dict[str, ndarray]], the cell-level NTScore, all niche-level NTScore dict,
    and all cell-level NTScore dict
    """

    info('Projecting NTScore from niche-level to cell-level.')

    cell_level_NTScore = np.zeros(niche_level_NTScore.shape[0])

    all_niche_level_NTScore_dict: Dict[str, ndarray] = {}
    all_cell_level_NTScore_dict: Dict[str, ndarray] = {}

    for i, data in enumerate(dataset):
        # the slice of data in each sample
        mask = data.mask
        slice_ = slice(i * data.x.shape[0], i * data.x.shape[0] + mask.sum())

        # niche to cell matrix
        niche_weight_matrix = load_npz(rel_params['Data'][i]['NicheWeightMatrix'])
        niche_to_cell_matrix = (
            niche_weight_matrix /
            niche_weight_matrix.sum(axis=0)).T  # normalize by the all niches associated with each cell, N x N

        # cell-level NTScore
        niche_level_NTScore_ = niche_level_NTScore[slice_].reshape(-1, 1)  # N x 1
        cell_level_NTScore_ = niche_to_cell_matrix @ niche_level_NTScore_
        cell_level_NTScore[slice_] = cell_level_NTScore_.reshape(-1)

        all_niche_level_NTScore_dict[data.name] = niche_level_NTScore_
        all_cell_level_NTScore_dict[data.name] = cell_level_NTScore_

    return cell_level_NTScore, all_niche_level_NTScore_dict, all_cell_level_NTScore_dict


def NTScore_table(options: Values, rel_params: Dict, all_niche_level_NTScore_dict: Dict[str, ndarray],
                  all_cell_level_NTScore_dict: Dict[str, ndarray]) -> None:
    """
    Generate NTScore table and save it
    :param options: Values, options
    :param rel_params: Dict, relative paths
    :param all_niche_level_NTScore_dict: Dict[str, ndarray], all niche-level NTScore dict
    :param all_cell_level_NTScore_dict: Dict[str, ndarray], all cell-level NTScore dict
    :return: pd.DataFrame, NTScore table
    """

    info('Output NTScore tables.')

    NTScore_table = pd.DataFrame()
    for sample in rel_params['Data']:
        coordinates_df = pd.read_csv(sample['Coordinates'], index_col=0)
        coordinates_df['Niche_NTScore'] = all_niche_level_NTScore_dict[sample['Name']]
        coordinates_df['Cell_NTScore'] = all_cell_level_NTScore_dict[sample['Name']]
        coordinates_df.to_csv(f'{options.NTScore_dir}/{sample["Name"]}_NTScore.csv.gz')
        NTScore_table = pd.concat([NTScore_table, coordinates_df])

    NTScore_table.to_csv(f'{options.NTScore_dir}/NTScore.csv.gz')



#!/usr/bin/env python

import sys
from typing import Callable, Optional

from ..model import GraphPooling
from ..optparser import opt_GP_validate, prepare_GP_optparser
from ..run.processes import NTScore, gnn, load_parameters
from ..train import GPBatchTrain
from ..train.inspect_funcs import loss_record
from ..utils import write_version_info
from ..niche_net import gen_original_data

# ------------------------------------
# Classes
# ------------------------------------


# ------------------------------------
# Functions
# ------------------------------------
def get_inspect_funcs() -> Optional[list[Callable]]:
    """
    Inspect function list
    :param output_dir: output dir
    :param epoch_filter: epoch filter
    :return: list of inspect functions
    """
    return [loss_record]


# ------------------------------------
# Main Function
# ------------------------------------
def main() -> None:
    """
    Main function
    :return: None
    """

    # write version information
    write_version_info()

    # load parameters
    options = load_parameters(opt_validate_func=opt_GP_validate, prepare_optparser_func=prepare_GP_optparser)

    # load original data
    ori_data_df = gen_original_data(options=options)

    # ----- GNN -----
    gnn(options=options,
        ori_data_df=ori_data_df,
        nn_model=GraphPooling,
        BatchTrain=GPBatchTrain,
        inspect_funcs=get_inspect_funcs())

    # ----- NT score -----
    NTScore(options=options)


# ------------------------------------
# Program running
# ------------------------------------
if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        sys.stderr.write("User interrupts me! ;-) See you ^.^!\n")
        sys.exit(0)



from optparse import OptionParser, Values

from ..log import *
from ..version import __version__
from ._IO import *
from ._train import *
from ._NT import add_NT_options_group, write_NT_options_memo

# ------------------------------------
# Constants
# ------------------------------------
IO_OPTIONS = ['input', 'preprocessing_dir', 'GNN_dir', 'NTScore_dir']


# ------------------------------------
# Functions
# ------------------------------------
def prepare_GP_optparser() -> OptionParser:
    """
    Prepare optparser object. New options will be added in thisfunction first.
    """
    usage = f'''USAGE: %prog <--meta-input META_INPUT> [--exp-input EXP_INPUT] [--embedding-input EMBEDDING_INPUT]
    [--decomposition-cell-type-composition-input DECOMPOSITION_CELL_TYPE_COMPOSITION_INPUT]
    [--decomposition-expression-input DECOMPOSITION_EXPRESSION_INPUT] <--preprocessing-dir PREPROCESSING_DIR>
    <--GNN-dir GNN_DIR> <--NTScore-dir NTSCORE_DIR>  [--device DEVICE] [--epochs EPOCHS] [--patience PATIENCE]
    [--min-delta MIN_DELTA] [--min-epochs MIN_EPOCHS] [--batch-size BATCH_SIZE] [-s SEED] [--seed SEED] [--lr LR]
    [--hidden-feats HIDDEN_FEATS] [-k K_CLUSTERS] [--modularity-loss-weight MODULARITY_LOSS_WEIGHT]
    [--purity-loss-weight PURITY_LOSS_WEIGHT] [--regularization-loss-weight REGULARIZATION_LOSS_WEIGHT] [--beta BETA]'''
    description = 'GP (Graph Pooling): GNN & Node Pooling'

    # option processor
    optparser = OptionParser(version=f'%prog {__version__}',
                             description=description,
                             usage=usage,
                             add_help_option=True)

    # I/O options group
    add_IO_options_group(optparser=optparser, io_options=IO_OPTIONS)

    # train and model options
    group_train = add_train_options_group(optparser)
    add_GNN_options_group(group_train)
    add_NP_options_group(group_train)

    # Niche trajectory
    add_NT_options_group(optparser)

    return optparser


def opt_GP_validate(optparser: OptionParser) -> Values:
    """Validate options from a OptParser object.

    Ret: Validated options object.
    """

    (options, args) = optparser.parse_args()

    validate_io_options(optparser, options, IO_OPTIONS)
    validate_train_options(optparser, options)

    info('------------------ RUN params memo ------------------ ')
    # print parameters to stdout
    write_io_options_memo(options, IO_OPTIONS)
    write_train_options_memo(options)
    write_GNN_options_memo(options)
    write_NP_options_memo(options)
    write_NT_options_memo(options)
    info('--------------- RUN params memo end ----------------- ')

    return options



from optparse import OptionParser, Values

from ..log import *
from ..version import __version__
from ._IO import *
from ._train import *

# ------------------------------------
# Constants
# ------------------------------------
IO_OPTIONS = ['preprocessing_dir', 'GNN_dir', 'NTScore_dir']


# ------------------------------------
# Functions
# ------------------------------------
def prepare_NT_optparser() -> OptionParser:
    """
    Prepare optparser object. New options will be added in thisfunction first.
    """
    usage = f'''USAGE: %prog <--preprocessing-dir PREPROCESSING_DIR> <--GNN-dir GNN_DIR> <--NTScore-dir NTSCORE_DIR>'''
    description = 'PseudoTime: Calculate PseudoTime for each node in a graph'

    # option processor
    optparser = OptionParser(version=f'%prog {__version__}', description=description, usage=usage, add_help_option=True)

    add_IO_options_group(optparser=optparser, io_options=IO_OPTIONS)

    return optparser


def add_NT_options_group(optparser: OptionParser) -> None:
    """
    Add niche trajectory options group to optparser.
    :param optparser: OptionParser object.
    :return: OptionGroup object.
    """

    group_NT = OptionGroup(optparser, "Options for niche trajectory")
    optparser.add_option_group(group_NT)
    group_NT.add_option(
        '--equal-space',
        dest='equal_space',
        action='store_true',
        default=False,
        help=
        'Whether to assign equally spaced values to for each niche cluster. Default is False, based on total loadings of each niche cluster.'
    )


def write_NT_options_memo(options: Values) -> None:
    """
    Write niche trajectory options memo.
    :param options: Values, options.
    :return: None.
    """

    info('---------------- Niche trajectory options ----------------')
    info(f'Equally spaced niche cluster scores: {options.equal_space}')


def opt_NT_validate(optparser: OptionParser) -> Values:
    """Validate options from a OptParser object.

    Ret: Validated options object.
    """

    (options, args) = optparser.parse_args()

    validate_io_options(optparser, options, IO_OPTIONS)

    # print parameters to stdout
    info('------------------ RUN params memo ------------------ ')
    write_io_options_memo(options, IO_OPTIONS)
    info('--------------- RUN params memo end ----------------- ')

    return options



import sys
from optparse import OptionGroup, OptionParser, Values

from ..log import *
from ..version import __version__
from ._IO import *

# ------------------------------------
# Constants
# ------------------------------------
IO_OPTIONS = ['input', 'preprocessing_dir']


# ------------------------------------
# Functions
# ------------------------------------
def add_preprocessing_options_group(optparser: OptionParser) -> None:
    """
    Add preprocessing options group to optparser.
    :param optparser: OptionParser object.
    :return: OptionGroup object.
    """
    # preprocessing options group
    group_preprocessing = OptionGroup(optparser, "Preprocessing")
    group_preprocessing.add_option(
        '--resolution',
        dest='resolution',
        type=float,
        default=1.0,
        help=
        'Resolution for leiden clustering. Used for clustering cells into cell types when gene expression data is provided. Default is 1.0.'
    )
    optparser.add_option_group(group_preprocessing)


def add_niche_net_constr_options_group(optparser: OptionParser) -> None:
    """
    Add niche network construction options group to optparser.
    :param optparser: OptionParser object.
    :return: OptionGroup object.
    """
    # niche network construction options group
    group_niche = OptionGroup(optparser, "Niche Network Construction")
    group_niche.add_option('--n-cpu',
                           dest='n_cpu',
                           type='int',
                           default=4,
                           help='Number of CPUs used for parallel computing in dataset preprocessing. Default is 4.')
    group_niche.add_option(
        '--n-neighbors',
        dest='n_neighbors',
        type='int',
        default=50,
        help=
        'Number of neighbors used for kNN graph construction. It should be less than the number of cells in each sample. Default is 50.'
    )
    group_niche.add_option(
        '--n-local',
        dest='n_local',
        type='int',
        default=20,
        help=
        'Specifies the nth closest local neighbors used for gaussian distance normalization. It should be less than the number of cells in each sample. Default is 20.'
    )
    group_niche.add_option(
        '--embedding-adjust',
        dest='embedding_adjust',
        action='store_true',
        default=False,
        help=
        'Adjust the cell type coding according to embeddings. Default is False. At least two (Embedding_1 and Embedding_2) should be in the original data if embedding_adjust is True.'
    )
    group_niche.add_option(
        '--sigma',
        dest='sigma',
        type='float',
        default=1,
        help=
        'Sigma for the exponential function to control the similarity between different cell types or clusters. Default is 1.'
    )
    optparser.add_option_group(group_niche)


def validate_niche_net_constr_options(optparser: OptionParser, options: Values) -> None:
    """
    Validate niche network construction options.
    :param optparser: OptionParser object.
    :param options: Options object.
    """
    if options.n_cpu < 1:
        error('n_cpu must be greater than 0.')
        optparser.print_help()
        sys.exit(1)

    if options.n_neighbors < 1:
        error('n_neighbors must be greater than 0.')
        optparser.print_help()
        sys.exit(1)

    if options.n_local < 1:
        error('n_local must be greater than 0.')
        optparser.print_help()
        sys.exit(1)

    if options.embedding_adjust:
        if options.embedding_input is None and options.exp_input is None and options.decomposition_expression_input is None:
            error('Please provide an embedding file or expression data file in csv format.')
            optparser.print_help()
            sys.exit(1)


def write_preprocessing_memo(options: Values):
    """Write preprocessing memos to stdout.

    Args:
        options: Options object.
    """

    # print parameters to stdout
    info('      -------- preprocessing options -------      ')
    info(f'resolution: {options.resolution}')

def write_niche_net_constr_memo(options: Values):
    """Write niche network construction memos to stdout.

    Args:
        options: Options object.
    """

    # print parameters to stdout
    info('      -------- niche net constr options -------      ')
    info(f'n_cpu:   {options.n_cpu}')
    info(f'n_neighbors: {options.n_neighbors}')
    info(f'n_local: {options.n_local}')
    info(f'embedding_adjust: {options.embedding_adjust}')
    info(f'sigma: {options.sigma}')


def prepare_preprocessing_optparser() -> OptionParser:
    """
    Prepare optparser object. New options will be added in thisfunction first.
    """

    usage = f'''USAGE: %prog <--meta-input META_INPUT> [--exp-input EXP_INPUT] [--embedding-input EMBEDDING_INPUT]
    [--decomposition-cell-type-composition-input DECOMPOSITION_CELL_TYPE_COMPOSITION_INPUT]
    [--decomposition-expression-input DECOMPOSITION_EXPRESSION_INPUT] <--preprocessing-dir PREPROCESSING_DIR> [--n-cpu N_CPU]
    [--n-neighbors N_NEIGHBORS] [--n-local N_LOCAL] [--embedding-adjust] [--sigma SIGMA]'''
    description = 'Create dataset for follwoing analysis.'

    # option processor
    optparser = OptionParser(version=f'%prog {__version__}', description=description, usage=usage, add_help_option=True)

    # I/O options group
    add_IO_options_group(optparser=optparser, io_options=IO_OPTIONS)

    # preprocessing options group
    add_preprocessing_options_group(optparser)

    # niche network construction options group
    add_niche_net_constr_options_group(optparser)

    return optparser


def opt_preprocessing_validate(optparser) -> Values:
    """Validate options from a OptParser object.

    Ret: Validated options object.
    """

    (options, args) = optparser.parse_args()

    validate_io_options(optparser=optparser, options=options, io_options=IO_OPTIONS)
    validate_niche_net_constr_options(optparser, options)

    # print parameters to stdout
    info('------------------ RUN params memo ------------------ ')
    write_io_options_memo(options, IO_OPTIONS)
    write_preprocessing_memo(options)
    write_niche_net_constr_memo(options)
    info('--------------- RUN params memo end ----------------- ')

    return options



import sys
from optparse import Values
from typing import List

import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix

from ..log import *


def perform_pca(expression_data: pd.DataFrame, n_components: int = 50) -> np.ndarray:
    """
    Perform PCA on the expression data.
    :param expression_data: Expression data.
    :param n_components: Number of components.
    :return: PCA result.
    """

    from sklearn.decomposition import PCA

    maximum_components = min(expression_data.shape[0], expression_data.shape[1])
    if n_components > maximum_components:
        n_components = maximum_components - 1

    model = PCA(n_components=n_components, svd_solver='arpack', random_state=0)
    info(f'Performing PCA with {n_components} components...')

    return model.fit_transform(X=expression_data.values)


def perform_harmony(embedding, meta_df: pd.DataFrame, batch_key: str) -> np.ndarray:
    """
    Perform Harmony on the embedding.
    :param embedding: Embedding.
    :param meta_df: Meta data.
    :param batch_key: Batch key.
    :return: Harmony result.
    """

    from harmonypy import run_harmony

    info(f'Performing Harmony using "{batch_key}" as batch ...')

    return run_harmony(data_mat=embedding, meta_data=meta_df, vars_use=batch_key)


def define_neighbors(embedding: np.ndarray, n_neighbors: int = 20) -> csr_matrix:
    """
    Define neighbors.
    :param embedding: Embedding.
    :param n_neighbors: Number of neighbors.
    :return: Neighbors.
    """

    #TODO: n_neighbors should be less than the number of cells in each sample.

    import pynndescent
    from scipy.sparse import coo_matrix
    from umap.umap_ import fuzzy_simplicial_set

    info(f'Defining neighbors with {n_neighbors} neighbors...')

    index = pynndescent.NNDescent(embedding, n_neighbors=n_neighbors, metric='euclidean', n_jobs=4)
    knn_indices, knn_dists = index.neighbor_graph
    connectivities, _, _, _ = fuzzy_simplicial_set(X=coo_matrix(([], ([], [])), shape=(embedding.shape[0], 1)),
                                                   n_neighbors=n_neighbors,
                                                   random_state=0,
                                                   metric="euclidean",
                                                   knn_indices=knn_indices,
                                                   knn_dists=knn_dists,
                                                   return_dists=False)

    return connectivities


def perform_umap(embedding: np.ndarray,
                 n_neighbors: int = 20,
                 min_dist: float = 0.5,
                 n_components: int = 2) -> np.ndarray:
    """
    Perform UMAP on the embedding.
    :param embedding: Embedding.
    :param n_neighbors: Number of neighbors.
    :param min_dist: Minimum distance.
    :param n_components: Number of components.
    :return: UMAP result.
    """

    import umap

    info(f'Performing UMAP with {n_neighbors} neighbors, {min_dist} minimum distance, and {n_components} components...')

    return umap.UMAP(n_neighbors=n_neighbors, min_dist=min_dist, n_components=n_components).fit_transform(embedding)


def perform_leiden(connectivities: csr_matrix, resolution: float = 1.0) -> List[int]:
    """
    Perform Leiden algorithm.
    :param connectivities: Connectivities.
    :param resolution: Resolution.
    :return: Leiden result.
    """

    import igraph as ig
    import leidenalg

    sources, targets = connectivities.nonzero()
    g = ig.Graph(directed=None)
    g.add_vertices(connectivities.shape[0])
    g.add_edges(es=list(zip(sources, targets)), attributes={'weight': connectivities[sources, targets].tolist()[0]})

    info(f'Performing Leiden algorithm with resolution {resolution}...')

    partition = leidenalg.find_partition(graph=g,
                                         partition_type=leidenalg.RBConfigurationVertexPartition,
                                         resolution_parameter=resolution)

    return partition.membership


def gen_original_data(options: Values) -> pd.DataFrame:
    """
    Load original data.

    option 1) there are gene expression data (csv format) and meta data (csv format)
    option 2) there are embedding data (csv format) and meta data (csv format)
    option 3) there is meta data (csv format) only
    option 4) there are decomposition input data (csv format) and meta data (csv format)
    """

    meta_data_df = pd.read_csv(options.meta_input, header=0, index_col=False, sep=',')

    if options.exp_input is None and options.embedding_input is None:
        if options.decomposition_cell_type_composition_input is None and options.decomposition_expression_input is None:  # option 3
            # check if Cell_Type column in the meta data
            if 'Cell_Type' not in meta_data_df.columns:
                raise ValueError(
                    'There are no expression data or embedding data. Please provide Cell_Type in the meta data.')
            meta_data_df['Cell_Type'] = meta_data_df['Cell_Type'].astype('category')
            # there should more than 1 cell type
            if len(meta_data_df['Cell_Type'].cat.categories) < 2:
                raise ValueError(
                    'There are no expression data or embedding data. Please provide at least two cell types in the meta data.'
                )
            # save mappings of the categorical data
            cell_type_code = pd.DataFrame(enumerate(meta_data_df['Cell_Type'].cat.categories),
                                          columns=['Code', 'Cell_Type'])
            cell_type_code.to_csv(f'{options.preprocessing_dir}/cell_type_code.csv', index=False)
            meta_data_df.to_csv(options.preprocessing_dir + '/meta_data.csv', index=False)
            return meta_data_df
        else:  # option 4
            if 'Cell_Type' in meta_data_df.columns:
                meta_data_df = meta_data_df.drop(columns=['Cell_Type'])
            decomposition_cell_type_composition_df = pd.read_csv(options.decomposition_cell_type_composition_input,
                                                                 header=0,
                                                                 index_col=0)
            decomposition_expression_df = pd.read_csv(options.decomposition_expression_input, header=0, index_col=0)
            if decomposition_cell_type_composition_df.shape[1] != decomposition_expression_df.shape[0]:
                error(
                    f'The number of cell type in the decomposition input data ({options.decomposition_cell_type_composition_input}) and ({options.decomposition_expression_input}) is not consistent. Please check it again.'
                )
                sys.exit(1)
            if decomposition_cell_type_composition_df.shape[0] != meta_data_df.shape[0]:
                warning(
                    f'The number of spots in the decomposition input data ({options.decomposition_cell_type_composition_input}) is not consistent with the meta data. We will use the intersection of them.'
                )
                common_spots = set(decomposition_cell_type_composition_df.index).intersection(
                    set(meta_data_df['Cell_ID']))
                decomposition_cell_type_composition_df = decomposition_cell_type_composition_df.loc[common_spots]
                meta_data_df = meta_data_df.loc[meta_data_df['Cell_ID'].isin(common_spots)]
            pca_embedding = perform_pca(decomposition_expression_df)
            if meta_data_df['Sample'].nunique() > 1:
                pca_embedding = perform_harmony(pca_embedding, meta_data_df, 'Sample')
            np.savetxt(options.preprocessing_dir + '/PCA_embedding.csv', pca_embedding, delimiter=',')
            for i in range(decomposition_cell_type_composition_df.shape[1]):
                meta_data_df[f'Cell_Type_{i}'] = decomposition_cell_type_composition_df.iloc[:, i]
            meta_data_df.to_csv(options.preprocessing_dir + '/meta_data.csv', index=False)
            cell_type_code = pd.DataFrame({
                'Code':
                range(decomposition_cell_type_composition_df.shape[1]),
                'Cell_Type': [f'Cell_Type_{i}' for i in range(decomposition_cell_type_composition_df.shape[1])]
            })
            cell_type_code.to_csv(f'{options.preprocessing_dir}/cell_type_code.csv', index=False)
            return meta_data_df

    if options.exp_input is None and options.embedding_input is not None:  # option 2
        meta_data_df['Cell_Type'] = meta_data_df['Cell_Type'].astype('category')
        # there should more than 1 cell type
        if len(meta_data_df['Cell_Type'].cat.categories) < 2:
            raise ValueError(
                'There are no expression data or embedding data. Please provide at least two cell types in the meta data.'
            )
        # save mappings of the categorical data
        cell_type_code = pd.DataFrame(enumerate(meta_data_df['Cell_Type'].cat.categories),
                                      columns=['Code', 'Cell_Type'])
        cell_type_code.to_csv(f'{options.preprocessing_dir}/cell_type_code.csv', index=False)

        embedding_df = pd.read_csv(options.embedding_input, header=0, index_col=0, sep=',')
        if embedding_df.shape[0] != meta_data_df.shape[0]:
            warning(
                f'The number of cells in the embedding data ({options.embedding_input}) is not consistent with the meta data. We will use the intersection of them.'
            )
            common_cells = set(embedding_df.index).intersection(set(meta_data_df['Cell_ID']))
            embedding_df = embedding_df.loc[common_cells]
            meta_data_df = meta_data_df.loc[meta_data_df['Cell_ID'].isin(common_spots)]
        for i in range(embedding_df.shape[1]):
            meta_data_df[f'Embedding_{i}'] = embedding_df.iloc[:, i]
        meta_data_df.to_csv(options.preprocessing_dir + '/meta_data.csv', index=False)
        return meta_data_df

    if options.exp_input is not None:  # option 1
        expression_data_df = pd.read_csv(options.exp_input, header=0, index_col=0, sep=',')
        if expression_data_df.shape[0] != meta_data_df.shape[0]:
            warning(
                f'The number of cells in the expression data ({options.exp_input}) is not consistent with the meta data. We will use the intersection of them.'
            )
            common_cells = set(expression_data_df.index).intersection(set(meta_data_df['Cell_ID']))
            expression_data_df = expression_data_df.loc[common_cells]
            meta_data_df = meta_data_df.loc[meta_data_df['Cell_ID'].isin(common_spots)]
        pca_embedding = perform_pca(expression_data_df)
        if 'Batch' in meta_data_df.columns:
            if meta_data_df['Batch'].nunique() > 1:
                pca_embedding = perform_harmony(pca_embedding, meta_data_df, 'Batch')
        elif meta_data_df['Sample'].nunique() > 1:
            pca_embedding = perform_harmony(pca_embedding, meta_data_df, 'Sample')
        np.savetxt(options.preprocessing_dir + '/PCA_embedding.csv', pca_embedding, delimiter=',')
        connectivities = define_neighbors(pca_embedding)
        leiden_result = perform_leiden(connectivities, resolution=options.resolution)
        umap_embedding = perform_umap(pca_embedding)
        np.savetxt(options.preprocessing_dir + '/UMAP_embedding.csv', umap_embedding, delimiter=',')
        meta_data_df['Cell_Type'] = pd.Categorical(leiden_result)
        for i in range(pca_embedding.shape[1]):
            meta_data_df[f'Embedding_{i}'] = pca_embedding[:, i]
        # save mappings of the categorical data
        cell_type_code = pd.DataFrame(enumerate(meta_data_df['Cell_Type'].cat.categories),
                                      columns=['Code', 'Cell_Type'])
        cell_type_code.to_csv(f'{options.preprocessing_dir}/cell_type_code.csv', index=False)
        meta_data_df.to_csv(options.preprocessing_dir + '/meta_data.csv', index=False)
        return meta_data_df



import torch
from torch import Tensor

from ..log import debug


def moran_I_features(X: Tensor, W: Tensor, mask: Tensor) -> Tensor:
    r"""
    Calculate Moran's I.
    :math:`I = \frac{n}{\sum_{i=1}^n\sum_{j=1}^n w_{ij}}\frac{(X-\bar{X})^T W (X-\bar{X})}{\sum_{i=1}^n (X_i-\bar{X})^2}`
    :param X: Tensor, shape: (N, F)
    :param W: Tensor, shape: (N, N)
    :param mask: Tensor, shape: (N, )
    :return: moran_I: Tensor, shape: (F, )
    """
    # --- input shape check ---
    X = X.unsqueeze(1) if X.dim() == 1 else X

    assert X.shape[0] == W.shape[0] == mask.shape[0]
    assert W.shape[0] == W.shape[1]

    F = X.shape[1]

    # --- mask ---
    X_masked = X[mask].reshape((-1, F))
    W_masked = W[mask][:, mask]
    # debug(f'X_masked shape: {X_masked.shape}.')
    # debug(f'W_masked shape: {W_masked.shape}.')

    # --- calculate ---
    n = X_masked.shape[0]
    X_masked_ = X_masked - X_masked.mean(dim=0, keepdim=True)
    W_masked_sum = W_masked.sum()
    # debug(f'X_masked_ shape: {X_masked_.shape}.')
    # debug(f'n: {n}.')
    # debug(f'W_masked_sum: {W_masked_sum}.')
    # debug(f'numerator: {torch.diagonal(X_masked_.T @ W_masked @ X_masked_)}.')  # F x 1
    # debug(f'denominator: {(X_masked_**2).sum(dim = 0, keepdim=True)}.')  # F x 1
    return n / W_masked_sum * torch.diagonal(X_masked_.T @ W_masked @ X_masked_) / (X_masked_**2).sum(dim=0,
                                                                                                      keepdim=True)


def graph_smooth_loss(z: Tensor, adj: Tensor, mask: Tensor) -> Tensor:
    r"""
    Graph smooth loss using -1 * moran's I
    :math: `L_{smooth} = \frac{1}{2}\sum_{i=1}^{N}\sum_{j=1}^{N}w_{ij}(z_i - z_j)^2`
    
    :param z: hidden embedding tensor
        :math:`\mathbf{Z} \in \mathbb{R}^{B \times N \times F}`, with
        batch-size :math:`B`, (maximum) number of nodes :math:`N` for
        each graph, and feature dimension :math:`F`.
    :param adj: adjacency tensor
        :math:`\mathbf{A} \in \mathbb{R}^{B \times N \times N}`.
    :param mask: mask tensor
        :math:`\mathbf{M} \in {\{ 0, 1 \}}^{B \times N}`
    :return: loss tensor
    """
    # --- inputs shape check ---
    z = z.unsqueeze(0) if z.dim() == 2 else z  # B x N x F
    adj = adj.unsqueeze(0) if adj.dim() == 2 else adj  # B x N x N
    mask = mask.unsqueeze(0) if mask.dim() == 1 else mask  # B x N
    B, N, F = z.size()

    # --- calculate moran's I ---
    MI_list = torch.cat(list(moran_I_features(z[i, :, :], adj[i], mask[i]) for i in range(B)))
    # debug(f'MI_list: {MI_list}')

    # --- calculate loss ---
    MI_loss = -1 * MI_list.mean()
    MI_loss = MI_loss + 1  # make the loss positive

    return MI_loss


def within_cluster_variance_loss(x: Tensor, s: Tensor, mask: Tensor) -> Tensor:
    """
    Calculate within cluster variance loss
    Args:
        x: input tensor, shape: (B, N, F)
        s: soft cluster assignment matrix, shape: (B, N, C)
        mask: mask tensor, shape: (B, N)
    Returns:
        loss: within cluster variance loss
    """

    # --- Extend mask to match the dimensions of x and s ---
    mask_extended_x = mask.unsqueeze(-1).expand_as(x)  # B x N x F
    mask_extended_s = mask.unsqueeze(-1).expand_as(s)  # B x N x C

    # --- Apply mask to x and s ---
    masked_x = x * mask_extended_x  # B x N x F
    masked_s = s * mask_extended_s  # B x N x C

    # --- Compute the cluster centroids ---
    sum_x = torch.einsum('bnf,bnc->cf', masked_x, masked_s)  # C x F, sum of each cluster
    num_points = torch.sum(masked_s, dim=(0, 1)) + 1e-10  # avoid divide by zero
    centroids = sum_x / num_points.unsqueeze(-1)  # C x F, mean of each cluster

    # --- Compute the squared distance from the centroids ---
    expanded_centroids = torch.einsum('cf,bnc->bnf', centroids, masked_s)  # B x N x F
    squared_distance = (masked_x - expanded_centroids)**2  # B x N x F

    # --- Compute the loss ---
    loss = torch.sum(squared_distance * mask_extended_x) / mask.sum()  # average over all nodes with mask

    return loss


def masked_variance(x, mask):
    """
    Args:
        x: input tensor, shape: (B, N, F)
        mask: mask tensor, shape: (B, N)
    Returns:
        variance
    """

    # Apply the mask
    x_masked = x * mask.unsqueeze(-1)

    # Calculate the mean
    sum_x = x_masked.sum()
    sum_mask = mask.sum() * x.shape[-1]
    mean_x = sum_x / sum_mask

    # Calculate the variance
    diff = x_masked - mean_x.unsqueeze(-1).unsqueeze(-1)
    var_x = torch.sum(diff * diff * mask.unsqueeze(-1)) / sum_mask

    return var_x



#!/usr/bin/env python

import sys

from ..optparser import opt_NT_validate, prepare_NT_optparser
from ..run.processes import NTScore, load_parameters
from ..utils import write_version_info


# ------------------------------------
# Main Function
# ------------------------------------
def main() -> None:
    """
    Main function
    :return: None
    """

    # write version information
    write_version_info()

    # load parameters
    options = load_parameters(opt_validate_func=opt_NT_validate, prepare_optparser_func=prepare_NT_optparser)

    # ----- NT score -----
    NTScore(options=options)


# ------------------------------------
# Program running
# ------------------------------------
if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        sys.stderr.write("User interrupts me! ;-) See you ^.^!\n")
        sys.exit(0)



from ._preprocessing import opt_preprocessing_validate, prepare_preprocessing_optparser
from ._GP import opt_GP_validate, prepare_GP_optparser
from ._NT import opt_NT_validate, prepare_NT_optparser
from ._ONTraC import opt_ontrac_validate, prepare_ontrac_optparser



import os
import sys
from optparse import OptionGroup, OptionParser, Values

from ..analysis.cell_type import cell_type_visualization
from ..analysis.data import AnaData
from ..analysis.niche_cluster import niche_cluster_visualization
from ..analysis.niche_net import embedding_adjust_visualization
from ..analysis.spatial import spatial_visualization
from ..analysis.train_loss import train_loss_visualiztion
from ..log import *
from ..optparser._IO import add_IO_options_group, validate_io_options
from ..utils import *
from ..version import __version__

# ------------------------------------
# Constants
# ------------------------------------
IO_OPTIONS = ['input', 'preprocessing_dir', 'GNN_dir', 'NTScore_dir']


# ------------------------------------
# Functions
# ------------------------------------
def analysis_pipeline(options: Values) -> None:
    # 0. load data class
    ana_data = AnaData(options)

    # part 1: niche network construction
    embedding_adjust_visualization(ana_data=ana_data)

    # part 2: train loss
    train_loss_visualiztion(ana_data=ana_data)

    # part 3: spatial based output
    spatial_visualization(ana_data=ana_data)

    # part 4: niche cluster
    niche_cluster_visualization(ana_data=ana_data)

    # part 5: cell type based output
    if not options.suppress_cell_type:
        cell_type_visualization(ana_data=ana_data)


# TODO: move to optparser
def add_suppress_group(optparser: OptionParser) -> None:
    group = OptionGroup(optparser, 'Suppress options')
    group.add_option('--suppress-cell-type-composition',
                     dest='suppress_cell_type_composition',
                     action='store_true',
                     default=False,
                     help='Suppress the cell type composition visualization.')
    group.add_option('--suppress-niche-cluster-loadings',
                     dest='suppress_niche_cluster_loadings',
                     action='store_true',
                     default=False,
                     help='Suppress the niche cluster loadings visualization.')
    group.add_option('--suppress-cell-type',
                     dest='suppress_cell_type',
                     action='store_true',
                     default=False,
                     help='Suppress the cell type visualization.')
    optparser.add_option_group(group)


def add_embedding_adjust_group(optparser: OptionParser) -> None:
    group = OptionGroup(optparser, 'Embedding adjust options')
    group.add_option(
        '--embedding-adjust',
        dest='embedding_adjust',
        action='store_true',
        default=False,
        help=
        'Adjust the cell type coding according to embeddings. Default is False. At least two (Embedding_1 and Embedding_2) should be in the original data if embedding_adjust is True.'
    )
    group.add_option(
        '--sigma',
        dest='sigma',
        type='float',
        default=1,
        help=
        'Sigma for the exponential function to control the similarity between different cell types or clusters. Default is 1.'
    )
    optparser.add_option_group(group)


def prepare_optparser() -> OptionParser:
    """Prepare optparser object. New options will be added in this function first.

    Ret: OptParser object.
    """
    usage = "USAGE: %prog <-d DATASET> <--preprocessing-dir PREPROCESSING_DIR> <--GNN-dir GNN_DIR> <--NTScore-dir NTSCORE_DIR> <-o OUTPUT_DIR> [-l LOG_FILE] [-r REVERSE]"
    description = "Analysis the results of ONTraC."
    optparser = OptionParser(version=f'%prog {__version__}',
                             usage=usage,
                             description=description,
                             add_help_option=False)
    optparser.add_option('-h', '--help', action='help', help='Show this help message and exit.')
    optparser.add_option('-o', '--output', dest='output', type='string', help='Output directory.')
    optparser.add_option('-l', '--log', dest='log', type='string', help='Log file.')
    optparser.add_option('-r',
                         '--reverse',
                         dest='reverse',
                         action='store_true',
                         default=False,
                         help='Reverse the NT score.')
    optparser.add_option('-s',
                         '--sample',
                         dest='sample',
                         action='store_true',
                         default=False,
                         help='Plot each sample separately.')
    add_IO_options_group(optparser=optparser, io_options=IO_OPTIONS)
    add_suppress_group(optparser)
    add_embedding_adjust_group(optparser)
    return optparser


def opt_validate(optparser: OptionParser) -> Values:
    """Validate options from a OptParser object.

    Args:
        optparser: OptParser object.
    """
    (options, args) = optparser.parse_args()

    validate_io_options(optparser, options, IO_OPTIONS, overwrite_validation=False)

    if not options.output:
        error('Output directory is required.')
        sys.exit(1)
    if not os.path.isdir(options.output):
        info(f'Output directory not found: {options.output}, will create it.')
        os.makedirs(options.output)
    else:
        warning(f'Output directory already exists: {options.output}, will overwrite it.')

    if not os.path.exists(options.log):
        error(f'Log file not found: {options.log}')
        sys.exit(1)

    options.yaml = f'{options.preprocessing_dir}/samples.yaml'
    if not os.path.exists(options.yaml):
        error(f'File not found: {options.yaml}')
        sys.exit(1)
    options.device = 'cpu'

    return options


def main():

    # write version information
    write_version_info()

    optparser = prepare_optparser()
    options = opt_validate(optparser)

    analysis_pipeline(options)


if __name__ == '__main__':
    main()



import os
import sys
from optparse import OptionGroup, OptionParser, Values
from typing import List, Optional

from ..log import *


def add_IO_options_group(optparser: OptionParser, io_options: Optional[List[str]]) -> None:
    """
    Add I/O options group to optparser.
    :param optparser: OptionParser object.
    :return: OptionGroup object.
    """
    if io_options is None:
        return
    # I/O options group
    group_io = OptionGroup(optparser, "IO")
    if 'input' in io_options:
        group_io.add_option(  # TODO: deprecated warning
            '-d',
            '--dataset',
            dest='dataset',
            type='string',
            help='Meta data file in csv format. Each row is a cell and each column is a meta data. The first column should be the cell name. Coordinates (x, y), cell type and sample should be included.'
        )
        group_io.add_option(
            '--meta-input',
            dest='meta_input',
            type='string',
            help=
            'Meta data file in csv format. Each row is a cell and each column is a meta data. The first column should be the cell name. Coordinates (x, y) and sample should be included. Cell type is optional.'
        )
        group_io.add_option(
            '--exp-input',
            dest='exp_input',
            type='string',
            help=
            'Normalized expression file in csv format. Each row is a cell and each column is a gene. The first column should be the cell name. If not provided, cell type should be included in the meta data file.'
        )
        group_io.add_option('--embedding-input',
                            dest='embedding_input',
                            type='string',
                            help='Embedding file in csv format. The first column should be the cell name.')
        group_io.add_option(
            '--decomposition-cell-type-composition-input',
            dest='decomposition_cell_type_composition_input',
            type='string',
            help=
            'Decomposition outputed cell type compostion of each spot in csv format. The first column should be the spot name.'
        )
        group_io.add_option(
            '--decomposition-expression-input',
            dest='decomposition_expression_input',
            type='string',
            help=
            'Decomposition outputed expression of each cell type in csv format. The first column should be the cell type name corresponding to the columns name of decomposition outputed cell type compostion.'
        )
    if 'preprocessing_dir' in io_options:
        group_io.add_option('--preprocessing-dir',
                            dest='preprocessing_dir',
                            type='string',
                            help='Directory for preprocessing outputs.')
    if 'GNN_dir' in io_options:
        group_io.add_option('--GNN-dir', dest='GNN_dir', type='string', help='Directory for the GNN output.')
    if 'NTScore_dir' in io_options:
        group_io.add_option('--NTScore-dir',
                            dest='NTScore_dir',
                            type='string',
                            help='Directory for the NTScore output.')

    optparser.add_option_group(group_io)


def validate_io_options(optparser: OptionParser,
                        options: Values,
                        io_options: Optional[List[str]],
                        overwrite_validation: bool = True) -> None:
    """Validate IO options from a OptParser object.
    :param optparser: OptionParser object.
    :param options: Values object.
    :param io_options: List of IO options.
    :param overwrite_validation: Overwrite validation.
    Ret: None
    """
    if io_options is None:
        return
    if 'input' in io_options:
        # dataset
        if options.dataset and not options.meta_input:
            warning(FutureWarning('The -d/--dataset option is deprecated. Please use --meta-input instead.').__str__())
            options.meta_input = options.dataset
        # meta data
        if not options.meta_input:
            error('Please provide a meta data file in csv format.')
            optparser.print_help()
            sys.exit(1)
        if not os.path.isfile(options.meta_input):
            error(f'The meta data file ({options.meta_input}) you given does not exist.')
            optparser.print_help()
            sys.exit(1)
        if not options.meta_input.endswith(('csv', 'csv.gz')):
            error(f'The meta data file ({options.meta_input}) should be in csv format.')
            optparser.print_help()
            sys.exit(1)
        # expression data
        if options.exp_input:
            if not os.path.isfile(options.exp_input):
                error(f'The expression data file ({options.exp_input}) you given does not exist.')
                optparser.print_help()
                sys.exit(1)
            if not options.exp_input.endswith(('csv', 'csv.gz')):
                error(f'The expression data file ({options.exp_input}) should be in csv format.')
                optparser.print_help()
                sys.exit(1)
        # embedding
        if options.embedding_input:
            if not os.path.isfile(options.embedding_input):
                error(f'The embedding file ({options.embedding_input}) you given does not exist.')
                optparser.print_help()
                sys.exit(1)
            if not options.embedding_input.endswith(('csv', 'csv.gz')):
                error(f'The embedding file ({options.embedding_input}) should be in csv format.')
                optparser.print_help()
                sys.exit(1)
        # decomposition
        # two decomposition input files should be provided together
        if options.decomposition_cell_type_composition_input and not options.decomposition_expression_input:
            error('Please provide both decomposition cell type compostion file and decomposition expression file.')
            optparser.print_help()
            sys.exit(1)
        if not options.decomposition_cell_type_composition_input and options.decomposition_expression_input:
            error('Please provide both decomposition cell type compostion file and decomposition expression file.')
            optparser.print_help()
            sys.exit(1)
        # check decomposition input files
        if options.decomposition_cell_type_composition_input:
            if not os.path.isfile(options.decomposition_cell_type_composition_input):
                error(
                    f'The decomposition outputed cell type compostion file ({options.decomposition_cell_type_composition_input}) you given does not exist.'
                )
                optparser.print_help()
                sys.exit(1)
            if not options.decomposition_cell_type_composition_input.endswith(('csv', 'csv.gz')):
                error(
                    f'The decomposition outputed cell type compostion file ({options.decomposition_cell_type_composition_input}) should be in csv format.'
                )
                optparser.print_help()
                sys.exit(1)
        if options.decomposition_expression_input:
            if not os.path.isfile(options.decomposition_expression_input):
                error(
                    f'The decomposition outputed expression file ({options.decomposition_expression_input}) you given does not exist.'
                )
                optparser.print_help()
                sys.exit(1)
            if not options.decomposition_expression_input.endswith(('csv', 'csv.gz')):
                error(
                    f'The decomposition outputed expression file ({options.decomposition_expression_input}) should be in csv format.'
                )
                optparser.print_help()

    if 'preprocessing_dir' in io_options:
        if not options.preprocessing_dir:
            error('Please provide a directory for preprocessing outputs.')
            optparser.print_help()
            sys.exit(1)
        if os.path.isdir(options.preprocessing_dir):
            if overwrite_validation:
                warning(
                    f'The directory ({options.preprocessing_dir}) you given already exists. It will be overwritten.')
            else:
                pass
        else:
            info(f'Creating directory: {options.preprocessing_dir}')
            os.makedirs(options.preprocessing_dir, exist_ok=True)

    if 'GNN_dir' in io_options:
        if not options.GNN_dir:
            error('Please provide a directory for the GNN output.')
            optparser.print_help()
            sys.exit(1)
        if os.path.isdir(options.GNN_dir):
            if overwrite_validation:
                warning(f'The directory ({options.GNN_dir}) you given already exists. It will be overwritten.')
            else:
                pass
        else:
            info(f'Creating directory: {options.GNN_dir}')
            os.makedirs(options.GNN_dir, exist_ok=True)

    if 'NTScore_dir' in io_options:
        if not options.NTScore_dir:
            error('Please provide a directory for the NTScore output.')
            optparser.print_help()
            sys.exit(1)
        if os.path.isdir(options.NTScore_dir):
            if overwrite_validation:
                warning(f'The directory ({options.NTScore_dir}) you given already exists. It will be overwritten.')
            else:
                pass
        else:
            info(f'Creating directory: {options.NTScore_dir}')
            os.makedirs(options.NTScore_dir, exist_ok=True)


def write_io_options_memo(options: Values, io_options: Optional[List[str]]) -> None:
    """Write IO options to stdout.

    Ret: None
    """
    if io_options is None:
        return
    info('            -------- I/O options -------             ')
    if 'input' in io_options:
        info(f'meta data file:  {options.meta_input}')
        if options.exp_input:
            info(f'expression data file:  {options.exp_input}')
        if options.embedding_input:
            info(f'embedding file:  {options.embedding_input}')
        if options.decomposition_cell_type_composition_input:
            info(f'decomposition cell type compostion file:  {options.decomposition_cell_type_composition_input}')
        if options.decomposition_expression_input:
            info(f'decomposition expression file:  {options.decomposition_expression_input}')
    if 'preprocessing_dir' in io_options:
        info(f'preprocessing output directory:  {options.preprocessing_dir}')
    if 'GNN_dir' in io_options:
        info(f'GNN output directory:  {options.GNN_dir}')
    if 'NTScore_dir' in io_options:
        info(f'NTScore output directory:  {options.NTScore_dir}')



import sys
from copy import deepcopy
from optparse import Values
from typing import Dict

import pandas as pd
import yaml

from ..log import warning


def write_version_info() -> None:
    """
    Write version information to stdout
    """
    from .. import __version__
    template = f'''##################################################################################

         ▄▄█▀▀██   ▀█▄   ▀█▀ █▀▀██▀▀█                   ▄▄█▀▀▀▄█
        ▄█▀    ██   █▀█   █     ██    ▄▄▄ ▄▄   ▄▄▄▄   ▄█▀     ▀
        ██      ██  █ ▀█▄ █     ██     ██▀ ▀▀ ▀▀ ▄██  ██
        ▀█▄     ██  █   ███     ██     ██     ▄█▀ ██  ▀█▄      ▄
         ▀▀█▄▄▄█▀  ▄█▄   ▀█    ▄██▄   ▄██▄    ▀█▄▄▀█▀  ▀▀█▄▄▄▄▀

                        version: {__version__}

##################################################################################
'''

    sys.stdout.write(template)
    sys.stdout.flush()


def load_meta_data(options: Values) -> pd.DataFrame:
    """
    Load meta data
    :param options: Values, options
    :return: pd.DataFrame, meta data

    1) read meta data file (csv format)
    2) check if Cell_ID, Sample, Cell_Type (optional), x, and y columns in the meta data
    3) make the Cell_Type column categorical
    4) return
        1. meta data with Cell_ID, Sample, Cell_Type (optional), x, and y columns
        2. samples
    """

    # read meta data file
    meta_data_df = pd.read_csv(options.dataset, header=0, index_col=False, sep=',')

    # check if Cell_ID, Sample, Cell_Type, x, and y columns in the meta data
    if 'Cell_ID' not in meta_data_df.columns:
        raise ValueError('Cell_ID column is missing in the meta data.')
    if 'Sample' not in meta_data_df.columns:
        raise ValueError('Sample column is missing in the meta data.')
    if 'x' not in meta_data_df.columns:
        raise ValueError('x column is missing in the meta data.')
    if 'y' not in meta_data_df.columns:
        raise ValueError('y column is missing in the meta data.')

    # check if there any duplicated Cell_ID
    if meta_data_df['Cell_ID'].duplicated().any():
        warning(
            'There are duplicated Cell_ID in the meta data. Sample name will added to Cell_ID to distinguish them.')
        meta_data_df['Cell_ID'] = meta_data_df['Sample'] + '_' + meta_data_df['Cell_ID']
    if meta_data_df['Cell_ID'].isnull().any():
        raise ValueError(
            f'Duplicated Cell_ID within same sample found! Please check the meta data file: {options.data_file}.')

    meta_data_df = meta_data_df.dropna()

    # make the Sample column string
    meta_data_df['Sample'] = meta_data_df['Sample'].astype(str)

    return meta_data_df


def read_yaml_file(yaml_file: str) -> dict:
    with open(yaml_file, 'r') as fhd:
        params = yaml.load(fhd, Loader=yaml.FullLoader)
    return params


def count_lines(filename: str) -> int:
    """
    Count lines of a file
    :param filename: file name
    :return: number of lines
    """
    i = 0
    if filename.endswith('.gz'):
        import gzip
        fhd = gzip.open(filename, 'rt')
    else:
        fhd = open(filename, 'r')
    for _ in fhd:
        i += 1
    fhd.close()
    return i


def get_rel_params(options: Values, params: Dict) -> Dict:
    """
    Get relative paths for params
    :param options: Values, options
    :param params: Dict, input samples
    :return: Dict, relative paths
    """
    rel_params = deepcopy(params)
    for data in rel_params['Data']:
        for k, v in data.items():
            if k == 'Name':
                continue
            data[k] = f'{options.preprocessing_dir}/{v}'
    return rel_params


def round_epoch_filter(epoch: int) -> bool:
    """
    Round epoch filter
    Only round epoch (1, 2, ..., 9, 10, 20, ..., 90, 100, ...) will be saved
    :param epoch: int
    :return: bool
    """

    def _is_power_of_10(n: int) -> bool:
        """
        Check if n is power of 10
        :param n: int
        :return: bool
        """
        num = len(str(n))
        return n % (10**(num - 1)) == 0

    return epoch < 10 or _is_power_of_10(epoch)



from typing import List, Optional, Tuple, Union

import matplotlib as mpl

mpl.rcParams['pdf.fonttype'] = 42
mpl.rcParams['ps.fonttype'] = 42
mpl.rcParams['font.family'] = 'Arial'
import matplotlib.pyplot as plt

from ..log import warning
from .data import AnaData


def plot_cell_type_composition_dataset(ana_data: AnaData) -> Optional[Tuple[plt.Figure, plt.Axes]]:
    """
    Plot spatial distribution of cell type composition.
    :param ana_data: AnaData, the data for analysis.
    :return: None or Tuple[plt.Figure, plt.Axes]
    """

    try:
        if ana_data.cell_type_composition is None or ana_data.cell_type_codes is None:
            warning("No cell type composition data found.")
            return None
    except FileNotFoundError as e:
        warning(str(e))
        return None

    samples: List[str] = ana_data.cell_type_composition['sample'].unique().tolist()
    cell_types: List[str] = ana_data.cell_type_codes['Cell_Type'].tolist()

    M, N = len(samples), len(cell_types)
    fig, axes = plt.subplots(M, N, figsize=(3.5 * N, 3 * M))
    for i, sample in enumerate(samples):
        sample_df = ana_data.cell_type_composition.loc[ana_data.cell_type_composition['sample'] == sample]
        for j, cell_type in enumerate(cell_types):
            ax = axes[i, j] if M > 1 else axes[j]
            scatter = ax.scatter(sample_df['x'],
                                 sample_df['y'],
                                 c=sample_df[cell_type],
                                 cmap='Reds',
                                 vmin=0,
                                 vmax=1,
                                 s=1)
            ax.set_xticks([])
            ax.set_yticks([])
            plt.colorbar(scatter)
            ax.set_title(f"{sample} {cell_type}")

    fig.tight_layout()
    if ana_data.options.output is not None:
        fig.savefig(f'{ana_data.options.output}/cell_type_composition.pdf', transparent=True)
        return None
    else:
        return fig, axes


def plot_cell_type_composition_sample(ana_data: AnaData) -> Optional[List[Tuple[plt.Figure, plt.Axes]]]:
    """
    Plot spatial distribution of cell type composition.
    :param ana_data: AnaData, the data for analysis.
    :return: None or Tuple[plt.Figure, plt.Axes]
    """

    try:
        if ana_data.cell_type_composition is None or ana_data.cell_type_codes is None:
            warning("No cell type composition data found.")
            return None
    except FileNotFoundError as e:
        warning(str(e))
        return None

    samples: List[str] = ana_data.cell_type_composition['sample'].unique().tolist()
    cell_types: List[str] = ana_data.cell_type_codes['Cell_Type'].tolist()

    output = []
    N = len(cell_types)
    for sample in samples:
        fig, axes = plt.subplots(1, N, figsize=(3.5 * N, 3))
        sample_df = ana_data.cell_type_composition.loc[ana_data.cell_type_composition['sample'] == sample]
        for j, cell_type in enumerate(cell_types):
            ax = axes[j]  # At least two cell types are required, checked at original data loading.
            scatter = ax.scatter(sample_df['x'],
                                 sample_df['y'],
                                 c=sample_df[cell_type],
                                 cmap='Reds',
                                 vmin=0,
                                 vmax=1,
                                 s=1)
            ax.set_xticks([])
            ax.set_yticks([])
            plt.colorbar(scatter)
            ax.set_title(f"{sample} {cell_type}")
        fig.tight_layout()
        if ana_data.options.output is not None:
            fig.savefig(f'{ana_data.options.output}/{sample}_cell_type_composition.pdf', transparent=True)
            plt.close(fig)
        else:
            output.append((fig, axes))

    return output if len(output) > 0 else None


def plot_cell_type_composition(
        ana_data: AnaData) -> Optional[Union[List[Tuple[plt.Figure, plt.Axes]], Tuple[plt.Figure, plt.Axes]]]:
    """
    Plot spatial distribution of cell type composition.
    :param ana_data: AnaData, the data for analysis.
    """

    if hasattr(ana_data.options, 'sample') and ana_data.options.sample:
        return plot_cell_type_composition_sample(ana_data=ana_data)
    else:
        return plot_cell_type_composition_dataset(ana_data=ana_data)


def plot_adjust_cell_type_composition_dataset(ana_data: AnaData) -> Optional[Tuple[plt.Figure, plt.Axes]]:
    """
    Plot spatial distribution of adjusted cell type composition.
    :param ana_data: AnaData, the data for analysis.
    :return: None or Tuple[plt.Figure, plt.Axes]
    """

    try:
        if ana_data.adjust_cell_type_composition is None or ana_data.cell_type_codes is None:
            warning("No adjusted cell type composition data found.")
            return None
    except FileNotFoundError as e:
        warning(str(e))
        return None

    samples: List[str] = ana_data.adjust_cell_type_composition['sample'].unique().tolist()
    cell_types: List[str] = ana_data.cell_type_codes['Cell_Type'].tolist()

    M, N = len(samples), len(cell_types)
    fig, axes = plt.subplots(M, N, figsize=(3.5 * N, 3 * M))
    for i, sample in enumerate(samples):
        sample_df = ana_data.adjust_cell_type_composition.loc[ana_data.adjust_cell_type_composition['sample'] == sample]
        for j, cell_type in enumerate(cell_types):
            ax = axes[i, j] if M > 1 else axes[j]
            scatter = ax.scatter(sample_df['x'],
                                 sample_df['y'],
                                 c=sample_df[cell_type],
                                 cmap='Reds',
                                 vmin=0,
                                 vmax=1,
                                 s=1)
            ax.set_xticks([])
            ax.set_yticks([])
            plt.colorbar(scatter)
            ax.set_title(f"{sample} {cell_type}")

    fig.tight_layout()
    if ana_data.options.output is not None:
        fig.savefig(f'{ana_data.options.output}/adjust_cell_type_composition.pdf', transparent=True)
        return None
    else:
        return fig, axes


def plot_adjust_cell_type_composition_sample(ana_data: AnaData) -> Optional[List[Tuple[plt.Figure, plt.Axes]]]:
    """
    Plot spatial distribution of adjusted cell type composition.
    :param ana_data: AnaData, the data for analysis.
    :return: None or Tuple[plt.Figure, plt.Axes]
    """

    try:
        if ana_data.adjust_cell_type_composition is None or ana_data.cell_type_codes is None:
            warning("No adjusted cell type composition data found.")
            return None
    except FileNotFoundError as e:
        warning(str(e))
        return None

    samples: List[str] = ana_data.adjust_cell_type_composition['sample'].unique().tolist()
    cell_types: List[str] = ana_data.cell_type_codes['Cell_Type'].tolist()

    output = []
    N = len(cell_types)
    for sample in samples:
        fig, axes = plt.subplots(1, N, figsize=(3.5 * N, 3))
        sample_df = ana_data.adjust_cell_type_composition.loc[ana_data.adjust_cell_type_composition['sample'] == sample]
        for j, cell_type in enumerate(cell_types):
            ax = axes[j]  # At least two cell types are required, checked at original data loading.
            scatter = ax.scatter(sample_df['x'],
                                 sample_df['y'],
                                 c=sample_df[cell_type],
                                 cmap='Reds',
                                 vmin=0,
                                 vmax=1,
                                 s=1)
            ax.set_xticks([])
            ax.set_yticks([])
            plt.colorbar(scatter)
            ax.set_title(f"{sample} {cell_type}")
        fig.tight_layout()
        if ana_data.options.output is not None:
            fig.savefig(f'{ana_data.options.output}/{sample}_adjust_cell_type_composition.pdf', transparent=True)
            plt.close(fig)
        else:
            output.append((fig, axes))

    return output if len(output) > 0 else None


def plot_adjust_cell_type_composition(
        ana_data: AnaData) -> Optional[Union[List[Tuple[plt.Figure, plt.Axes]], Tuple[plt.Figure, plt.Axes]]]:
    """
    Plot spatial distribution of adjusted cell type composition.
    :param ana_data: AnaData, the data for analysis.
    """

    if hasattr(ana_data.options, 'sample') and ana_data.options.sample:
        return plot_adjust_cell_type_composition_sample(ana_data=ana_data)
    else:
        return plot_adjust_cell_type_composition_dataset(ana_data=ana_data)


def plot_niche_NT_score_dataset(ana_data: AnaData) -> Optional[Tuple[plt.Figure, plt.Axes]]:
    """
    Plot spatial distribution of niche NT score.
    :param ana_data: AnaData, the data for analysis.
    :param NT_score_reverse: bool, whether to reverse the NT score.
    :return: None or Tuple[plt.Figure, plt.Axes]
    """

    samples: List[str] = ana_data.NT_score['sample'].unique().tolist()

    try:
        if 'Niche_NTScore' not in ana_data.NT_score.columns:
            warning("Niche_NTScore not found in the NT score data.")
            return None
    except FileNotFoundError as e:
        warning(str(e))
        return None

    N = len(samples)
    fig, axes = plt.subplots(1, N, figsize=(3.5 * N, 3))
    for i, sample in enumerate(samples):
        sample_df = ana_data.NT_score.loc[ana_data.NT_score['sample'] == sample]
        ax = axes[i] if N > 1 else axes
        NT_score = sample_df['Niche_NTScore'] if not ana_data.options.reverse else 1 - sample_df['Niche_NTScore']
        scatter = ax.scatter(sample_df['x'], sample_df['y'], c=NT_score, cmap='rainbow', vmin=0, vmax=1, s=1)
        ax.set_xticks([])
        ax.set_yticks([])
        plt.colorbar(scatter)
        ax.set_title(f"{sample} Niche-level NT Score")

    fig.tight_layout()
    if ana_data.options.output is not None:
        fig.savefig(f'{ana_data.options.output}/niche_NT_score.pdf', transparent=True)
        return None
    else:
        return fig, axes


def plot_niche_NT_score_sample(ana_data: AnaData) -> Optional[List[Tuple[plt.Figure, plt.Axes]]]:
    """
    Plot spatial distribution of niche NT score.
    :param ana_data: AnaData, the data for analysis.
    :param NT_score_reverse: bool, whether to reverse the NT score.
    :return: None or Tuple[plt.Figure, plt.Axes]
    """

    samples: List[str] = ana_data.NT_score['sample'].unique().tolist()

    try:
        if 'Niche_NTScore' not in ana_data.NT_score.columns:
            warning("Niche_NTScore not found in the NT score data.")
            return None
    except FileNotFoundError as e:
        warning(str(e))
        return None

    output = []
    for sample in samples:
        fig, ax = plt.subplots(1, 1, figsize=(3.5, 3))
        sample_df = ana_data.NT_score.loc[ana_data.NT_score['sample'] == sample]
        NT_score = sample_df['Niche_NTScore'] if not ana_data.options.reverse else 1 - sample_df['Niche_NTScore']
        scatter = ax.scatter(sample_df['x'], sample_df['y'], c=NT_score, cmap='rainbow', vmin=0, vmax=1, s=1)
        ax.set_xticks([])
        ax.set_yticks([])
        plt.colorbar(scatter)
        ax.set_title(f"{sample} Niche-level NT Score")
        fig.tight_layout()
        if ana_data.options.output is not None:
            fig.savefig(f'{ana_data.options.output}/{sample}_niche_NT_score.pdf', transparent=True)
            plt.close(fig)
        else:
            output.append((fig, ax))

    return output if len(output) > 0 else None


def plot_niche_NT_score(
        ana_data: AnaData) -> Optional[Union[List[Tuple[plt.Figure, plt.Axes]], Tuple[plt.Figure, plt.Axes]]]:
    """
    Plot spatial distribution of niche NT score.
    :param ana_data: AnaData, the data for analysis.
    """

    if hasattr(ana_data.options, 'sample') and ana_data.options.sample:
        return plot_niche_NT_score_sample(ana_data=ana_data)
    else:
        return plot_niche_NT_score_dataset(ana_data=ana_data)


def plot_cell_NT_score_dataset(ana_data: AnaData) -> Optional[Tuple[plt.Figure, plt.Axes]]:
    """
    Plot spatial distribution of cell NT score.
    :param ana_data: AnaData, the data for analysis.
    :param NT_score_reverse: bool, whether to reverse the NT score.
    :return: None or Tuple[plt.Figure, plt.Axes]
    """

    samples: List[str] = ana_data.NT_score['sample'].unique().tolist()

    try:
        if 'Cell_NTScore' not in ana_data.NT_score.columns:
            warning("Cell_NTScore not found in the NT score data.")
            return None
    except FileNotFoundError as e:
        warning(str(e))
        return None

    N = len(samples)
    fig, axes = plt.subplots(1, N, figsize=(3.5 * N, 3))
    for i, sample in enumerate(samples):
        sample_df = ana_data.NT_score.loc[ana_data.NT_score['sample'] == sample]
        ax = axes[i] if N > 1 else axes
        NT_score = sample_df['Cell_NTScore'] if not ana_data.options.reverse else 1 - sample_df['Cell_NTScore']
        scatter = ax.scatter(sample_df['x'], sample_df['y'], c=NT_score, cmap='rainbow', vmin=0, vmax=1, s=1)
        ax.set_xticks([])
        ax.set_yticks([])
        plt.colorbar(scatter)
        ax.set_title(f"{sample} Cell-level NT Score")

    fig.tight_layout()
    if ana_data.options.output is not None:
        fig.savefig(f'{ana_data.options.output}/cell_NT_score.pdf', transparent=True)
        return None
    else:
        return fig, axes


def plot_cell_NT_score_sample(ana_data: AnaData) -> Optional[List[Tuple[plt.Figure, plt.Axes]]]:
    """
    Plot spatial distribution of cell NT score.
    :param ana_data: AnaData, the data for analysis.
    :param NT_score_reverse: bool, whether to reverse the NT score.
    :return: None or Tuple[plt.Figure, plt.Axes]
    """

    samples: List[str] = ana_data.NT_score['sample'].unique().tolist()

    try:
        if 'Cell_NTScore' not in ana_data.NT_score.columns:
            warning("Cell_NTScore not found in the NT score data.")
            return None
    except FileNotFoundError as e:
        warning(str(e))
        return None

    output = []
    for sample in samples:
        fig, ax = plt.subplots(1, 1, figsize=(3.5, 3))
        sample_df = ana_data.NT_score.loc[ana_data.NT_score['sample'] == sample]
        NT_score = sample_df['Cell_NTScore'] if not ana_data.options.reverse else 1 - sample_df['Cell_NTScore']
        scatter = ax.scatter(sample_df['x'], sample_df['y'], c=NT_score, cmap='rainbow', vmin=0, vmax=1, s=1)
        ax.set_xticks([])
        ax.set_yticks([])
        plt.colorbar(scatter)
        ax.set_title(f"{sample} Cell-level NT Score")
        fig.tight_layout()
        if ana_data.options.output is not None:
            fig.savefig(f'{ana_data.options.output}/{sample}_cell_NT_score.pdf', transparent=True)
            plt.close(fig)
        else:
            output.append((fig, ax))

    return output if len(output) > 0 else None


def plot_cell_NT_score(
        ana_data: AnaData) -> Optional[Union[List[Tuple[plt.Figure, plt.Axes]], Tuple[plt.Figure, plt.Axes]]]:
    """
    Plot spatial distribution of cell NT score.
    :param ana_data: AnaData, the data for analysis.
    """

    if hasattr(ana_data.options, 'sample') and ana_data.options.sample:
        return plot_cell_NT_score_sample(ana_data=ana_data)
    else:
        return plot_cell_NT_score_dataset(ana_data=ana_data)


def spatial_visualization(ana_data: AnaData) -> None:
    """
    All spatial visualization will include here.
    :param ana_data: AnaData, the data for analysis.
    """

    # 1. cell type compostion
    if not hasattr(ana_data.options,
                   'suppress_cell_type_composition') or not ana_data.options.suppress_cell_type_composition:
        plot_cell_type_composition(ana_data=ana_data)
        if hasattr(ana_data.options, 'embedding_adjust') and ana_data.options.embedding_adjust:
            plot_adjust_cell_type_composition(ana_data=ana_data)

    # 2. NT score
    plot_niche_NT_score(ana_data=ana_data)
    plot_cell_NT_score(ana_data=ana_data)



from typing import Optional, Tuple

import matplotlib as mpl

mpl.rcParams['pdf.fonttype'] = 42
mpl.rcParams['ps.fonttype'] = 42
mpl.rcParams['font.family'] = 'Arial'
import matplotlib.pyplot as plt
import seaborn as sns

from .data import AnaData


def train_loss_visualiztion(ana_data: AnaData) -> Optional[Tuple[plt.Figure, plt.Axes]]:
    if ana_data.train_loss is None:
        return None
    train_loss_data = ana_data.train_loss
    loss_df, final_loss_dict = train_loss_data['loss_df'], train_loss_data['loss_dict']

    fig, axes = plt.subplots(loss_df.shape[1], 1, figsize=(6.4, 2 * loss_df.shape[1]))
    for index, loss_name in enumerate(loss_df.columns):
        sns.lineplot(data=loss_df, x='Epoch', y=loss_name, ax=axes[index])
        if loss_name in final_loss_dict:
            axes[index].annotate(f'Final: {final_loss_dict[loss_name]:.4f}',
                                 xy=(0.98, 0.98),
                                 xycoords='axes fraction',
                                 ha='right',
                                 va='top',
                                 fontsize=12)
            axes[index].set_ylabel(loss_name.replace('_', ' ').capitalize())
    fig.tight_layout()
    if ana_data.options.output is not None:
        fig.savefig(f'{ana_data.options.output}/train_loss.pdf', dpi=300)
        plt.close(fig)
        return None
    else:
        return fig, axes



from optparse import Values
from typing import Callable, List, Optional, Type

import numpy as np
import pandas as pd
import torch

from ..GNN import (evaluate, load_data, predict, save_graph_pooling_results,
                   set_seed, train)
from ..log import *
from ..niche_net import (construct_niche_network, ct_coding_adjust,
                         gen_original_data, gen_samples_yaml)
from ..niche_trajectory import (NTScore_table, get_niche_NTScore,
                                load_consolidate_data, niche_to_cell_NTScore)
from ..train import SubBatchTrainProtocol
from ..utils import get_rel_params, read_yaml_file


def load_parameters(opt_validate_func: Callable, prepare_optparser_func: Callable) -> Values:
    """
    Load parameters
    :param opt_validate_func: validate function
    :param prepare_optparser_func: prepare optparser function
    :return: options
    """
    options = opt_validate_func(prepare_optparser_func())

    return options


def niche_network_construct(options: Values, ori_data_df: pd.DataFrame) -> None:
    """
    Niche network construct process
    :param options: options
    :param ori_data_df: pd.DataFrame, original data
    :return: None
    """

    info('------------- Niche network construct --------------- ')
    ori_data_df = gen_original_data(options=options)

    # construct niche network
    construct_niche_network(options=options, ori_data_df=ori_data_df)

    # generate samples.yaml to indicate file paths for each sample
    gen_samples_yaml(options=options, ori_data_df=ori_data_df)

    # cell type coding adjust
    if options.embedding_adjust:
        ct_coding_adjust(options=options, ori_data_df=ori_data_df)

    info('------------ Niche network construct end ------------ ')


def gnn(options: Values, ori_data_df: pd.DataFrame, nn_model: Type[torch.nn.Module],
        BatchTrain: Type[SubBatchTrainProtocol], inspect_funcs: Optional[List[Callable]]) -> None:
    """
    GNN training and prediction process
    :param options: options
    :param ori_data_df: pd.DataFrame, original data
    :return: None
    """

    info('------------------------ GNN ------------------------ ')
    # load data
    dataset, sample_loader = load_data(options=options)
    # random seed
    set_seed(seed=options.seed)
    # build model
    input_feats = dataset.num_features
    model = nn_model(input_feats=input_feats, hidden_feats=options.hidden_feats, k=options.k, exponent=options.beta)
    # train
    batch_train = train(options=options,
                        nn_model=model,
                        BatchTrain=BatchTrain,
                        sample_loader=sample_loader,
                        inspect_funcs=inspect_funcs)
    # evaluate
    evaluate(batch_train=batch_train, model_name='GraphPooling')
    # predict
    consolidate_s_array, _ = predict(output_dir=options.GNN_dir,
                                     batch_train=batch_train,
                                     dataset=dataset,
                                     model_name='GraphPooling')
    # save results
    if consolidate_s_array is not None:
        save_graph_pooling_results(ori_data_df=ori_data_df,
                                   dataset=dataset,
                                   rel_params=get_rel_params(
                                       options=options,
                                       params=read_yaml_file(f'{options.preprocessing_dir}/samples.yaml')),
                                   consolidate_s_array=consolidate_s_array,
                                   output_dir=options.GNN_dir)
    info('--------------------- GNN end ---------------------- ')


def NTScore(options: Values) -> None:
    """
    Pseudotime calculateion process
    :param options: options
    :param consolidate_s_array: consolidate s array
    :param consolidate_out_adj_array: consolidate out adj array
    :return: None
    """

    info('----------------- Niche trajectory ------------------ ')
    consolidate_s_array, consolidate_out_adj_array = load_consolidate_data(options=options)

    params = read_yaml_file(f'{options.preprocessing_dir}/samples.yaml')
    rel_params = get_rel_params(options, params)
    dataset, _ = load_data(options=options)

    niche_cluster_score, niche_level_NTScore = get_niche_NTScore(options=options,
                                                                 niche_cluster_loading=consolidate_s_array,
                                                                 niche_adj_matrix=consolidate_out_adj_array)
    cell_level_NTScore, all_niche_level_NTScore_dict, all_cell_level_NTScore_dict = niche_to_cell_NTScore(
        dataset=dataset, rel_params=rel_params, niche_level_NTScore=niche_level_NTScore)

    np.savetxt(fname=f'{options.NTScore_dir}/niche_cluster_score.csv.gz', X=niche_cluster_score, delimiter=',')
    np.savetxt(fname=f'{options.NTScore_dir}/niche_NTScore.csv.gz', X=niche_level_NTScore, delimiter=',')
    np.savetxt(fname=f'{options.NTScore_dir}/cell_NTScore.csv.gz', X=cell_level_NTScore, delimiter=',')

    NTScore_table(options=options,
                  rel_params=rel_params,
                  all_niche_level_NTScore_dict=all_niche_level_NTScore_dict,
                  all_cell_level_NTScore_dict=all_cell_level_NTScore_dict)
    info('--------------- Niche trajectory end ---------------- ')



from typing import Optional, Tuple, Union

import matplotlib as mpl
import numpy as np
import pandas as pd
from scipy.spatial import distance

mpl.rcParams['pdf.fonttype'] = 42
mpl.rcParams['ps.fonttype'] = 42
mpl.rcParams['font.family'] = 'Arial'
import matplotlib.pyplot as plt
import seaborn as sns

from ..log import info, warning
from ..niche_net import get_embedding_columns
from .data import AnaData


def clustering_visualization(ana_data: AnaData) -> Optional[Union[Tuple, None]]:
    """Visualization of clustering results.
    
    Args:
        ana_data: AnaData object.
    """

    if not ana_data.options.decomposition_expression_input:
        return None

    # load meta data
    meta_data = pd.read_csv(f'{ana_data.options.preprocessing_dir}/meta_data.csv', index_col=0)
    umap_embedding = np.loadtxt(f'{ana_data.options.preprocessing_dir}/PCA_embedding.csv', delimiter=',')
    data_df = meta_data['Cell_Type']
    data_df['Embedding_1'] = umap_embedding[:, 0]
    data_df['Embedding_2'] = umap_embedding[:, 1]
    with sns.axes_style('white', rc={
            'xtick.bottom': True,
            'ytick.left': True
    }), sns.plotting_context('paper',
                             rc={
                                 'axes.titlesize': 8,
                                 'axes.labelsize': 8,
                                 'xtick.labelsize': 6,
                                 'ytick.labelsize': 6,
                                 'legend.fontsize': 6
                             }):
        fig, ax = plt.subplots(1, 1, figsize=(7, 5))
        sns.scatterplot(data=data_df, x='Embedding_1', y='Embedding_2', hue='Cell_Type', s=2, ax=ax)
        ax.set_xlabel('UMAP_1')
        ax.set_ylabel('UMAP_2')
        ax.legend(loc='upper left', bbox_to_anchor=(1, 1), ncol=3, markerscale=4)
        if ana_data.options.output:
            fig.savefig(f'{ana_data.options.output}/clustering.pdf', transparent=True)
            return None
        else:
            return fig


def embedding_adjust_visualization(ana_data: AnaData) -> Optional[Union[Tuple, None]]:
    """Visualization of embedding adjust.

    Args:
        ana_data: AnaData object.
    """

    if not ana_data.options.embedding_adjust:
        return None

    # load original data
    ori_data_df = pd.read_csv(f'{ana_data.options.preprocessing_dir}/meta_data.csv', index_col=0)
    ct_coding = pd.read_csv(f'{ana_data.options.preprocessing_dir}/cell_type_code.csv', index_col=0)
    if ana_data.options.decomposition_expression_input:
        ct_embedding = np.loadtxt(f'{ana_data.options.preprocessing_dir}/PCA_embedding.csv', delimiter=',')
        raw_distance = distance.cdist(ct_embedding, ct_embedding, 'euclidean')
    else:
        embedding_columns = get_embedding_columns(ori_data_df)
        if len(embedding_columns) < 2:
            warning('At least two (Embedding_1 and Embedding_2) should be in the original data. Skip the adjustment.')
            return None

        # calculate embedding postion for each cell type
        ct_embedding = ori_data_df[embedding_columns + ['Cell_Type']].groupby('Cell_Type').mean()
        raw_distance = distance.cdist(ct_embedding[embedding_columns].values, ct_embedding[embedding_columns].values,
                                      'euclidean')

    # calculate distance between each cell type
    median_distance = np.median(raw_distance[np.triu_indices(raw_distance.shape[0], k=1)])
    info(f'Median distance between cell types: {median_distance}')
    raw_distance_df = pd.DataFrame(raw_distance, index=ct_coding['Cell_Type'], columns=ct_coding['Cell_Type'])

    if ana_data.options.sigma is None:
        ana_data.options.sigma = np.median(raw_distance[np.triu_indices(raw_distance.shape[0], k=1)])
        info(
            f'Sigma is not provided. Use the median ({ana_data.options.sigma}) of the distances between the cell type pairs.'
        )

    # calculate the M
    M = np.exp(-(raw_distance / (ana_data.options.sigma * median_distance))**2)
    M_df = pd.DataFrame(M, index=ct_coding['Cell_Type'], columns=ct_coding['Cell_Type'])

    with sns.axes_style('white', rc={
            'xtick.bottom': True,
            'ytick.left': True
    }), sns.plotting_context('paper',
                             rc={
                                 'axes.titlesize': 8,
                                 'axes.labelsize': 8,
                                 'xtick.labelsize': 6,
                                 'ytick.labelsize': 6,
                                 'legend.fontsize': 6
                             }):
        dis_cluster_grid = sns.clustermap(raw_distance_df,
                                          figsize=(raw_distance_df.shape[0] / 6, raw_distance_df.shape[0] / 6))
        M_cluster_grid = sns.clustermap(M_df, figsize=(M_df.shape[0] / 6, M_df.shape[0] / 6), vmin=0, vmax=1)
        if ana_data.options.output:
            dis_cluster_grid.savefig(f'{ana_data.options.output}/raw_distance.pdf', transparent=True)
            M_cluster_grid.savefig(f'{ana_data.options.output}/M.pdf', transparent=True)
            return None
        else:
            return dis_cluster_grid, M_cluster_grid
